# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

from collections import defaultdict
import logging
import weakref
from enum import Enum

from ....known_identities import PLATFORM_WEB
from .base import SubsystemBase

__docformat__ = "reStructuredText"

_log = logging.getLogger(__name__)


class ResourceType(Enum):
    JSONRPC = "jsonrpc"
    ENDPOINT = "endpoint"
    RAW = "raw"
    UNKNOWN = 99


class WebSubSystem(SubsystemBase):
    """
    The web subsystem handles the agent side of routing web data from the
    :class:`volttron.client.web.PlatformWebService`.

    """

    def __init__(self, owner, core, rpc):
        self._owner = weakref.ref(owner)
        self._rpc = weakref.ref(rpc)
        self._core = weakref.ref(core)
        self._endpoints = {}
        self._ws_endpoint = {}

        def onsetup(sender, **kwargs):
            rpc.export(self._opened, "client.opened")
            rpc.export(self._closed, "client.closed")
            rpc.export(self._message, "client.message")
            rpc.export(self._route_callback, "route.callback")

        def onstop(sender, **kwargs):
            rpc.call(PLATFORM_WEB, "unregister_all_agent_routes")

        core.onstop.connect(onstop, self)
        core.onsetup.connect(onsetup, self)

    def get_user_claims(self, bearer):
        return self._rpc().call(PLATFORM_WEB, "get_user_claims", bearer).get(timeout=10)

    def unregister_all_routes(self):
        self._rpc().call(PLATFORM_WEB, "unregister_all_agent_routes").get(timeout=10)

    def register_endpoint(self, endpoint, callback, res_type: ResourceType = ResourceType.JSONRPC):
        """
        The :meth:`register_endpoint` method registers an endpoint with the
        :param res_type:
        :class:`volttron.client.web.PlatformWebService` on the VOLTTRON
        instance.

        Each endpoint can map to at most one callback function.  The callback
        function must support the following interface

        .. code-block:: python

            def callback(self, env, data):
                print('The environmental variables {}'.format(env))
                print('The data sent {}'.format(data))

        .. versionadded:: VOLTTRON 4.0.1

        :param endpoint:
            Http endpoint matching the PATH_INFO environmental variable
        :param callback: Agent method to be called with the env and data.
        :type endpoint: str
        :type callback: function
        """
        _log.info("Registering route endpoint: {}".format(endpoint))
        self._endpoints[endpoint] = callback
        if isinstance(res_type, ResourceType):
            res_type = res_type.value
        self._rpc().call(PLATFORM_WEB, "register_endpoint", endpoint, res_type).get(timeout=10)

    def register_path(self, prefix, static_path):
        """
        The :meth:`register_path` method registers a prefix that can be used
        for routing static files.

        .. versionadded:: VOLTTRON 4.0.1

        :param prefix:
        :param static_path:
            An existing path available to the
            :class:`volttron.client.web.PlatformWebService`
        :type prefix: str
        :type static_path: str
        """
        _log.info("Registering path prefix: {}, path: {}".format(prefix, static_path))
        self._rpc().call(PLATFORM_WEB, "register_path_route", prefix, static_path).get(timeout=10)

    def register_websocket(self, endpoint, opened=None, closed=None, received=None):
        """
        The :meth:`register_websocket` method registers a websocket endpoint
        that can be connected to through the
        :class:`volttron.client.web.PlatformWebService`.

        The parameters opened and closed can be specified as callback events
        with the following signature:

        .. code-block:: python

            def ws_opened(self, endpoint):
                print('ws_opened endpoint {}'.format(endpoint))

            def ws_closed(self, endpoint):
                print('ws_closed endpoint {}'.format(endpoint))

        The received event is triggered when the websocket is written to from
        the client.  The received event must have a signature such as the
        following interface:

        .. code-block:: python

            def ws_received(self, endpoint, message):
                print('ws_received endpoint {} message: {}'.format(endpoint,
                                                                   message))

        .. versionadded:: VOLTTRON 4.0.1

        :param endpoint: The endpoint of the websocket event occurred on.
        :param opened:
            An event triggered when a client is connected to the endpoint.
        :param closed:
            An event triggered when a client is closed or disconnected from
            the endpoint.
        :param received:
            An event triggered when data comes in on the endpoint's websocket.
        :type endpoint: str
        :type opened: function
        :type closed: function
        :type received: function
        """
        _log.info("Agent registering websocket at: {}".format(endpoint))
        self._ws_endpoint[endpoint] = (opened, closed, received)
        self._rpc().call(PLATFORM_WEB, "register_websocket", endpoint).get(timeout=5)

    def unregister_websocket(self, endpoint):
        self._rpc().call(PLATFORM_WEB, "unregister_websocket", endpoint).get(timeout=5)

    def send(self, endpoint, message=""):
        """
        The :meth:`send` method publishes data to the registered websocket
        clients that are subscribed to the passed endpoint.

        .. versionadded:: VOLTTRON 4.0.1

        :param endpoint: The endpoint to be used to send the message.
        :param message:
            The message to be sent through to the client.  This parameter must
            be serializable.
        :type endpoint: str
        :type message: str
        """
        self._rpc().call(PLATFORM_WEB, "websocket_send", endpoint, message).get(timeout=5)

    def _route_callback(self, env, data):
        fn = self._endpoints.get(env["PATH_INFO"])
        if fn:
            _log.debug("Calling function: {}".format(fn.__name__))
            return fn(env, data)

        return None

    def _opened(self, fromip, endpoint):
        _log.info("Client opened websocket ip: {} endpoint: {}".format(fromip, endpoint))
        callbacks = self._ws_endpoint.get(endpoint)
        if callbacks is None:
            _log.error("Websocket endpoint {} is not available".format(endpoint))
        else:
            if callbacks[0]:
                return callbacks[0](fromip, endpoint)

        return False

    def _closed(self, endpoint):
        _log.info("Client closed websocket endpoint: {}".format(endpoint))

        callbacks = self._ws_endpoint.get(endpoint)
        if callbacks is None:
            _log.error("Websocket endpoint {} is not available".format(endpoint))
        else:
            if callbacks[1]:
                callbacks[1](endpoint)

    def _message(self, endpoint, message):

        callbacks = self._ws_endpoint.get(endpoint)
        if callbacks is None:
            _log.error("Websocket endpoint {} is not available".format(endpoint))
        else:
            if callbacks[2]:
                callbacks[2](endpoint, message)



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
import logging
import traceback
import weakref
import fnmatch
import inspect

from .base import SubsystemBase
from volttron.utils.storeutils import list_unique_links, check_for_config_link

# from volttron.client.storeutils import list_unique_links, check_for_config_link
from volttron.client.vip.agent import errors
from volttron.client.known_identities import CONFIGURATION_STORE

from collections import defaultdict
from copy import deepcopy
"""The configstore subsystem manages the agent side of the configuration store.
It is responsible for processing change notifications from the platform
 and triggering the correct callbacks with the contents of a configuration.
"""

__docformat__ = "reStructuredText"
__version__ = "1.0"

_log = logging.getLogger(__name__)

VALID_ACTIONS = set(["NEW", "UPDATE", "DELETE"])


class ConfigStore(SubsystemBase):

    def __init__(self, owner, core, rpc):
        self._core = weakref.ref(core)
        self._rpc = weakref.ref(rpc)

        self._ref_map = {}    # For triggering callbacks.
        self._reverse_ref_map = defaultdict(set)    # For triggering callbacks.
        self._store = {}
        self._default_store = {}
        self._callbacks = {}
        self._name_map = {}
        self._default_name_map = {}

        self._initialized = False
        self._initial_callbacks_called = False

        self._process_callbacks_code_object = self._process_callbacks.__code__

        def sub_factory():
            return defaultdict(set)

        self._subscriptions = defaultdict(sub_factory)

        def onsetup(sender, **kwargs):
            rpc.export(self._update_config, "config.update")
            rpc.export(self._initial_update, "config.initial_update")

        core.onsetup.connect(onsetup, self)
        core.configuration.connect(self._onconfig, self)

    def _onconfig(self, sender, **kwargs):
        if not self._initialized:
            try:
                self._rpc().call(CONFIGURATION_STORE, "get_configs").get()
            except errors.Unreachable as e:
                _log.error("Connected platform does not support the Configuration Store feature.")
                return
            except errors.VIPError as e:
                _log.error("Error retrieving agent configurations: {}".format(e))
                return

        affected_configs = {}
        for config_name in self._store:
            affected_configs[config_name] = "NEW"
        for config_name in self._default_store:
            affected_configs[config_name] = "NEW"

        self._process_callbacks(affected_configs)
        self._initial_callbacks_called = True

    def _add_refs(self, config_name, contents):
        refs = list_unique_links(contents)
        self._ref_map[config_name] = refs

        for ref in refs:
            self._reverse_ref_map[ref].add(config_name)

    def _update_refs(self, config_name, contents):
        self._delete_refs(config_name)

        self._add_refs(config_name, contents)

    def _delete_refs(self, config_name):
        # Delete refs if they exist.
        old_refs = self._ref_map.pop(config_name, set())

        for ref in old_refs:
            reverse_ref_set = self._reverse_ref_map[ref]
            reverse_ref_set.remove(config_name)
            if not reverse_ref_set:
                del self._reverse_ref_map[ref]

    def _initial_update(self, configs, reset_name_map=True):
        self._initialized = True
        self._store = {key.lower(): value for (key, value) in configs.items()}
        if reset_name_map:
            self._name_map = {key.lower(): key for key in configs}

        for config_name, config_contents in self._store.items():
            self._add_refs(config_name, config_contents)

        for config_name, config_contents in self._default_store.items():
            if config_name not in self._store:
                self._add_refs(config_name, config_contents)

    def _process_links(self, config_contents, already_gathered):
        if isinstance(config_contents, dict):
            for key, value in config_contents.items():
                if isinstance(value, (dict, list)):
                    self._process_links(value, already_gathered)
                elif isinstance(value, str):
                    config_name = check_for_config_link(value)
                    if config_name is not None:
                        config_contents[key] = self._gather_child_configs(
                            config_name, already_gathered)
        elif isinstance(config_contents, list):
            for i, value in enumerate(config_contents):
                if isinstance(value, (dict, list)):
                    self._process_links(value, already_gathered)
                elif isinstance(value, str):
                    config_name = check_for_config_link(value)
                    if config_name is not None:
                        config_contents[i] = self._gather_child_configs(
                            config_name, already_gathered)

    def _gather_child_configs(self, config_name, already_gathered):
        if config_name in already_gathered:
            return already_gathered[config_name]

        config_contents = self._store.get(config_name)
        if config_contents is None:
            config_contents = self._default_store.get(config_name)

        config_contents = deepcopy(config_contents)
        already_gathered[config_name] = config_contents

        self._process_links(config_contents, already_gathered)

        return config_contents

    def _gather_config(self, config_name):
        config_contents = self._store.get(config_name)
        if config_contents is None:
            config_contents = self._default_store.get(config_name)

        if config_contents is None:
            raise KeyError("{} not in store".format(config_name))

        already_configured = {}

        return self._gather_child_configs(config_name, already_configured)

    def _gather_affected(self, config_name, seen_dict):
        reverse_refs = self._reverse_ref_map[config_name]
        for ref in reverse_refs:
            if ref not in seen_dict:
                seen_dict[ref] = "UPDATE"
                self._gather_affected(ref, seen_dict)

    def _update_config(self, action, config_name, contents=None, trigger_callback=False):
        """Called by the platform to push out configuration changes."""
        # If we haven't yet grabbed the initial callback state we just bail.
        if not self._initialized:
            return

        affected_configs = {}

        # Update local store.
        if action == "DELETE":
            config_name_lower = config_name.lower()
            if config_name_lower in self._store:
                del self._store[config_name_lower]

                if config_name_lower not in self._default_store:
                    affected_configs[config_name_lower] = "DELETE"
                    self._gather_affected(config_name_lower, affected_configs)
                    self._delete_refs(config_name_lower)
                else:
                    affected_configs[config_name_lower] = "UPDATE"
                    self._gather_affected(config_name_lower, affected_configs)
                    self._update_refs(config_name_lower, self._default_store[config_name_lower])

        if action == "DELETE_ALL":
            for name in self._store:
                affected_configs[name] = "DELETE"
            # Just assume all default stores updated.
            for name in self._default_store:
                affected_configs[name] = "UPDATE"
            self._ref_map = {}
            self._reverse_ref_map = defaultdict(set)
            self._initial_update({}, False)

        if action in ("NEW", "UPDATE"):
            config_name_lower = config_name.lower()
            self._store[config_name_lower] = contents
            self._name_map[config_name_lower] = config_name
            if config_name_lower in self._default_store:
                action = "UPDATE"
            affected_configs[config_name_lower] = action
            self._update_refs(config_name_lower, self._store[config_name_lower])
            self._gather_affected(config_name_lower, affected_configs)

        if trigger_callback and self._initial_callbacks_called:
            self._process_callbacks(affected_configs)

        if action == "DELETE":
            del self._name_map[config_name_lower]

        if action == "DELETE_ALL":
            self._name_map.clear()

    def _process_callbacks(self, affected_configs):
        _log.debug("Processing callbacks for affected files: {}".format(affected_configs))
        all_map = self._default_name_map.copy()
        all_map.update(self._name_map)
        # Always process "config" first.
        if "config" in affected_configs:
            self._process_callbacks_one_config("config", affected_configs["config"], all_map)

        for config_name, action in affected_configs.items():
            if config_name == "config":
                continue
            self._process_callbacks_one_config(config_name, action, all_map)

    def _process_callbacks_one_config(self, config_name, action, name_map):
        callbacks = set()
        for pattern, actions in self._subscriptions.items():
            if fnmatch.fnmatchcase(config_name, pattern) and action in actions:
                callbacks.update(actions[action])

        for callback in callbacks:
            try:
                if action == "DELETE":
                    contents = None
                else:
                    contents = self._gather_config(config_name)
                callback(name_map[config_name], action, contents)
            except Exception:
                tb_str = traceback.format_exc()
                _log.error("Problem processing callback:")
                _log.error(tb_str)

    def list(self):
        """Returns a list of configuration names for this agent.

        :returns: Configuration names
        :rtype: list

        :Return Values:
        A list of all the configuration names available for this agent.
        """
        # Handle case were we are called during "onstart".
        if not self._initialized:
            try:
                self._rpc().call(CONFIGURATION_STORE, "get_configs").get()
            except errors.Unreachable as e:
                _log.error("Connected platform does not support the Configuration Store feature.")
            except errors.VIPError as e:
                _log.error("Error retrieving agent configurations: {}".format(e))

        all_map = self._default_name_map.copy()
        all_map.update(self._name_map)

        store_set = set(self._store.keys())
        default_set = set(self._default_store.keys())
        config_list = list(all_map[x] for x in (store_set | default_set))
        config_list.sort()
        return config_list

    def get(self, config_name="config"):
        """Returns the contents of a configuration.

        :param config_name: Name of configuration to add to store.
        :type config_name: str
        :returns: Configuration contents
        :rtype: dict, list, or string

        :Return Values:
        The contents of the configuration specified.
        """
        # Handle case were we are called during "onstart".

        # If we fail to initialize we don't raise an exception as there still
        # may be a default configuration to grab.
        if not self._initialized:
            try:
                self._rpc().call(CONFIGURATION_STORE, "get_configs").get()
            except errors.Unreachable as e:
                _log.error("Connected platform does not support the Configuration Store feature.")
            except errors.VIPError as e:
                _log.error("Error retrieving agent configurations: {}".format(e))

        config_name = config_name.lower()

        return self._gather_config(config_name)

    def _check_call_from_process_callbacks(self):
        frame_records = inspect.stack()
        try:
            # Don't create any unneeded references to frame objects.
            for frame, *_ in frame_records:
                if self._process_callbacks_code_object is frame.f_code:
                    raise RuntimeError(
                        "Cannot request changes to the config store from a configuration callback."
                    )
        finally:
            del frame_records

    def set(self, config_name, contents, trigger_callback=False, send_update=True):
        """Called to set the contents of a configuration.

        May not be called before the onstart phase of an agents lifetime.

        May not be called from a configuration callback. Will produce a runtime error if done so.

        :param config_name: Name of configuration to add to store.
        :param contents: Contents of the configuration. May be a string, dictionary, or list.
        :param trigger_callback: Tell the platform to trigger callbacks on the agent for this change.

        :type config_name: str
        :type contents: str, dict, list
        :type trigger_callback: bool
        """
        self._check_call_from_process_callbacks()

        self._rpc().call(
            CONFIGURATION_STORE,
            "set_config",
            config_name,
            contents,
            trigger_callback=trigger_callback,
            send_update=send_update,
        ).get(timeout=10.0)

    def set_default(self, config_name, contents):
        """Called to set the contents of a default configuration file. Default configurations are used if the
        configuration store does not contain a configuration with that name.

        May not be called after the onsetup phase of an agents lifetime. Will produce a runtime error if done so.

        :param config_name: Name of configuration to add to store.
        :param contents: Contents of the configuration. May be a string, dictionary, or list.
        :type config_name: str
        :type contents: str, dict, list
        """
        if self._initialized:
            raise RuntimeError("Cannot request changes to default configurations after onsetup.")

        if not isinstance(contents, (str, list, dict)):
            raise ValueError("Invalid content type: {}".format(contents.__class__.__name__))

        config_name_lower = config_name.lower()
        self._default_store[config_name_lower] = contents
        self._default_name_map[config_name_lower] = config_name

        if config_name_lower in self._store:
            return

        self._update_refs(config_name_lower, self._default_store[config_name_lower])

    def delete_default(self, config_name):
        """Called to delete the contents of a default configuration file.

        May not be called after the onsetup phase of an agents lifetime. Will produce a runtime error if done so.

        :param config_name: Configuration to remove from store.
        :type config_name: str
        """
        if self._initialized:
            raise RuntimeError("Cannot request changes to default configurations after onsetup.")

        config_name_lower = config_name.lower()
        del self._default_store[config_name_lower]
        del self._default_name_map[config_name_lower]

        if config_name_lower in self._store:
            return

        self._update_refs(config_name_lower, self._store[config_name_lower])

    def delete(self, config_name, trigger_callback=False, send_update=True):
        """Delete a configuration by name. May not be called from a callback as this will cause
            deadlock with the platform. Will produce a runtime error if done so.

        :param config_name: Configuration to remove from store.
        :param trigger_callback: Tell the platform to trigger callbacks on the agent for this change.
        :type config_name: str
        :type trigger_callback: bool
        """
        self._check_call_from_process_callbacks()

        self._rpc().call(
            CONFIGURATION_STORE,
            "delete_config",
            config_name,
            trigger_callback=trigger_callback,
            send_update=send_update,
        ).get(timeout=10.0)

    def subscribe(self, callback, actions=VALID_ACTIONS, pattern="*"):
        """Subscribe to changes to a configuration.

        :param callback: Function to call in response to changes to a configuration.
        :param actions: Change actions to respond to. Valid values are "NEW", "UPDATE", and "DELETE". May be a single action or a list of actions.
        :param pattern: Configuration name pattern to match to.  Uses Unix style filename pattern matching.

        :type callback: str
        :type actions: str or list
        :type pattern: str
        """
        if isinstance(actions, str):
            actions = (actions, )

        actions = set(action.upper() for action in actions)

        invalid_actions = actions - VALID_ACTIONS
        if invalid_actions:
            raise ValueError("Invalid actions: " + list(invalid_actions))

        pattern = pattern.lower()

        for action in actions:
            self._subscriptions[pattern][action].add(callback)

    def unsubscribe_all(self):
        """Remove all subscriptions."""
        self._subscriptions.clear()



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging
import glob
import os
import os.path
import errno
from csv import DictReader
from io import StringIO

import gevent

from gevent.lock import Semaphore

from volttron.types import ServiceInterface
from volttron.utils import (
    parse_json_config,
    get_aware_utc_now,
    format_timestamp,
)
from volttron.utils import jsonapi

from volttron.utils.persistance import PersistentDict

from volttron.utils.jsonrpc import RemoteError, MethodNotFound
from volttron.utils.storeutils import check_for_recursion, strip_config_name, store_ext
from volttron.client.vip.agent import Agent, Core, RPC, Unreachable, VIPError

_log = logging.getLogger(__name__)

UPDATE_TIMEOUT = 30.0


def process_store(identity, store):
    """Parses raw store data and returns contents.
    Called at startup to initialize the parsed version of the store."""
    results = {}
    name_map = {}
    sync_store = False
    for config_name, config_data in store.items():
        config_type = config_data["type"]
        config_string = config_data["data"]
        try:
            processed_config = process_raw_config(config_string, config_type)
            if check_for_recursion(config_name, processed_config, results):
                raise ValueError("Recursive configuration references")
            results[config_name] = processed_config
        except ValueError as e:
            _log.error("Error processing Agent {} config {}: {}".format(
                identity, config_name, str(e)))
            sync_store = True
            del store[config_name]

        if config_name.lower() in name_map:
            _log.error("Conflicting names in store, dropping {}".format(config_name))
            sync_store = True
            del store[config_name]

        else:
            name_map[config_name.lower()] = config_name

    if sync_store:
        _log.warning("Removing invalid configurations for Agent {}".format(identity))
        store.sync()

    return results, name_map


def process_raw_config(config_string, config_type="raw"):
    """Parses raw config string into python objects"""
    if config_type == "raw":
        return config_string
    elif config_type == "json":
        config = parse_json_config(config_string)
        if not isinstance(config, (list, dict)):
            raise ValueError("Configuration must be a list or object.")
        return config
    elif config_type == "csv":
        f = StringIO(config_string)
        return [x for x in DictReader(f)]

    raise ValueError("Unsupported configuration type.")


class ConfigStoreService(ServiceInterface):

    def __init__(self, **kwargs):
        super(ConfigStoreService, self).__init__(**kwargs)

        # This agent is started before the router so we need
        # to keep it from blocking.
        self.core.delay_running_event_set = False

        self.store = {}
        self.store_path = os.path.join(os.environ["VOLTTRON_HOME"], "configuration_store")

    @Core.receiver("onsetup")
    def _setup(self, sender, **kwargs):
        _log.info("Initializing configuration store service.")

        try:
            # explicitly provide access to others. Needed for secure mode.
            # Agents needs access to its own store file in this dir
            os.makedirs(self.store_path)
            os.chmod(self.store_path, 0o755)
        except OSError as e:
            if e.errno != errno.EEXIST:
                _log.critical("Failed to create configuration store directory: " + str(e))
                raise
            else:
                _log.debug("Configuration directory already exists.")

        config_store_iter = glob.iglob(os.path.join(self.store_path, "*" + store_ext))

        for store_path in config_store_iter:
            root, ext = os.path.splitext(store_path)
            agent_identity = os.path.basename(root)
            _log.debug("Processing store for agent {}".format(agent_identity))
            store = PersistentDict(filename=store_path, flag="c", format="json")
            parsed_configs, name_map = process_store(agent_identity, store)
            self.store[agent_identity] = {
                "configs": parsed_configs,
                "store": store,
                "name_map": name_map,
                "lock": Semaphore(),
            }

    @RPC.export
    @RPC.allow("edit_config_store")
    def manage_store(self, identity, config_name, raw_contents, config_type="raw"):
        contents = process_raw_config(raw_contents, config_type)
        self._add_config_to_store(
            identity,
            config_name,
            raw_contents,
            contents,
            config_type,
            trigger_callback=True,
        )

    @RPC.export
    @RPC.allow("edit_config_store")
    def manage_delete_config(self, identity, config_name):
        self.delete(identity, config_name, trigger_callback=True)

    @RPC.export
    @RPC.allow("edit_config_store")
    def manage_delete_store(self, identity):
        agent_store = self.store.get(identity)
        if agent_store is None:
            return

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        agent_configs.clear()
        agent_disk_store.clear()
        agent_name_map.clear()

        # Sync will delete the file if the store is empty.
        agent_disk_store.async_sync()

        with agent_store_lock:
            try:
                self.vip.rpc.call(identity,
                                  "config.update",
                                  "DELETE_ALL",
                                  None,
                                  trigger_callback=True).get(timeout=UPDATE_TIMEOUT)
            except Unreachable:
                _log.debug("Agent {} not currently running. Configuration update not sent.".format(
                    identity))
            except RemoteError as e:
                _log.error("Agent {} failure when all configurations: {}".format(identity, e))
            except MethodNotFound as e:
                _log.error("Agent {} failure when deleting configuration store: {}".format(
                    identity, e))

        # If the store is still empty (nothing jumped in and added to it while
        # we were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    @RPC.export
    def manage_list_configs(self, identity):
        result = list(self.store.get(identity, {}).get("store", {}).keys())
        result.sort()
        return result

    @RPC.export
    def manage_list_stores(self):
        result = list(self.store.keys())
        result.sort()
        return result

    @RPC.export
    def manage_get(self, identity, config_name, raw=True):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        if raw:
            return agent_disk_store[real_config_name]["data"]

        return agent_configs[real_config_name]

    @RPC.export
    def manage_get_metadata(self, identity, config_name):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        agent_disk_store = agent_store["store"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        real_config = agent_disk_store[real_config_name]

        # Set modified to none if we predate the modified flag.
        if real_config.get("modified") is None:
            real_config["modified"] = None

        return real_config

    @RPC.export
    def set_config(self, config_name, contents, trigger_callback=False, send_update=True):
        identity = self.vip.rpc.context.vip_message.peer
        self.store_config(
            identity,
            config_name,
            contents,
            trigger_callback=trigger_callback,
            send_update=send_update,
        )

    @RPC.export
    def get_configs(self):
        """
        Called by an Agent at startup to trigger initial configuration state
        push.
        """
        identity = self.vip.rpc.context.vip_message.peer

        # We need to create store and lock if it doesn't exist in case someone
        # tries to add a configuration while we are sending the initial state.
        agent_store = self.store.get(identity)

        if agent_store is None:
            # Initialize a new store.
            store_path = os.path.join(self.store_path, identity + store_ext)
            store = PersistentDict(filename=store_path, flag="c", format="json")
            agent_store = {
                "configs": {},
                "store": store,
                "name_map": {},
                "lock": Semaphore(),
            }
            self.store[identity] = agent_store

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]

        with agent_store_lock:
            try:
                self.vip.rpc.call(identity, "config.initial_update",
                                  agent_configs).get(timeout=UPDATE_TIMEOUT)
            except Unreachable:
                _log.debug("Agent {} not currently running. Configuration update not sent.".format(
                    identity))
            except RemoteError as e:
                _log.error("Agent {} failure when performing initial update: {}".format(
                    identity, e))
            except MethodNotFound as e:
                _log.error("Agent {} failure when performing initial update: {}".format(
                    identity, e))
            except VIPError as e:
                _log.error("VIP Error sending initial agent configuration: {}".format(e))

        # If the store is empty (and nothing jumped in and added to it while we
        # were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    @RPC.export
    def delete_config(self, config_name, trigger_callback=False, send_update=True):
        """Called by an Agent to delete a configuration."""
        identity = self.vip.rpc.context.vip_message.peer
        self.delete(
            identity,
            config_name,
            trigger_callback=trigger_callback,
            send_update=send_update,
        )

    # Helper method to allow the local services to delete configs before message
    # bus in online.
    def delete(self, identity, config_name, trigger_callback=False, send_update=True):
        agent_store = self.store.get(identity)
        if agent_store is None:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            raise KeyError('No configuration file "{}" for VIP IDENTIY {}'.format(
                config_name, identity))

        real_config_name = agent_name_map[config_name_lower]

        agent_configs.pop(real_config_name)
        agent_disk_store.pop(real_config_name)
        agent_name_map.pop(config_name_lower)

        # Sync will delete the file if the store is empty.
        agent_disk_store.async_sync()

        if send_update:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(
                        identity,
                        "config.update",
                        "DELETE",
                        config_name,
                        trigger_callback=trigger_callback,
                    ).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when deleting configuration {}: {}".format(
                        identity, config_name, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))

        # If the store is empty (and nothing jumped in and added to it while we
        # were informing the agent) then remove it from the global store.
        if not agent_disk_store:
            self.store.pop(identity, None)

    # Helper method to allow the local services to store configs before message
    # bus is online.
    def store_config(self,
                     identity,
                     config_name,
                     contents,
                     trigger_callback=False,
                     send_update=True):
        config_type = None
        raw_data = None
        if isinstance(contents, (dict, list)):
            config_type = "json"
            raw_data = jsonapi.dumps(contents)
        elif isinstance(contents, str):
            config_type = "raw"
            raw_data = contents
        else:
            raise ValueError("Unsupported configuration content type: {}".format(
                str(type(contents))))

        self._add_config_to_store(
            identity,
            config_name,
            raw_data,
            contents,
            config_type,
            trigger_callback=trigger_callback,
            send_update=send_update,
        )

    def _add_config_to_store(
        self,
        identity,
        config_name,
        raw,
        parsed,
        config_type,
        trigger_callback=False,
        send_update=True,
    ):
        """Adds a processed configuration to the store."""
        agent_store = self.store.get(identity)

        action = "UPDATE"

        if agent_store is None:
            # Initialize a new store.
            store_path = os.path.join(self.store_path, identity + store_ext)
            store = PersistentDict(filename=store_path, flag="c", format="json")
            agent_store = {
                "configs": {},
                "store": store,
                "name_map": {},
                "lock": Semaphore(),
            }
            self.store[identity] = agent_store

        agent_configs = agent_store["configs"]
        agent_disk_store = agent_store["store"]
        agent_store_lock = agent_store["lock"]
        agent_name_map = agent_store["name_map"]

        config_name = strip_config_name(config_name)
        config_name_lower = config_name.lower()

        if config_name_lower not in agent_name_map:
            action = "NEW"

        if check_for_recursion(config_name, parsed, agent_configs):
            raise ValueError("Recursive configuration references detected.")

        if config_name_lower in agent_name_map:
            old_config_name = agent_name_map[config_name_lower]
            del agent_configs[old_config_name]

        agent_configs[config_name] = parsed
        agent_name_map[config_name_lower] = config_name

        agent_disk_store[config_name] = {
            "type": config_type,
            "modified": format_timestamp(get_aware_utc_now()),
            "data": raw,
        }

        agent_disk_store.async_sync()

        _log.debug("Agent {} config {} stored.".format(identity, config_name))

        if send_update:
            with agent_store_lock:
                try:
                    self.vip.rpc.call(
                        identity,
                        "config.update",
                        action,
                        config_name,
                        contents=parsed,
                        trigger_callback=trigger_callback,
                    ).get(timeout=UPDATE_TIMEOUT)
                except Unreachable:
                    _log.debug(
                        "Agent {} not currently running. Configuration update not sent.".format(
                            identity))
                except RemoteError as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))
                except MethodNotFound as e:
                    _log.error("Agent {} failure when adding/updating configuration {}: {}".format(
                        identity, config_name, e))
                except gevent.timeout.Timeout:
                    _log.error("Config update to agent {} timed out after {} seconds".format(
                        identity, UPDATE_TIMEOUT))
                except Exception as e:
                    _log.error("Unknown error sending update to agent identity {}.: {}".format(
                        identity, e))



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import random
from weakref import WeakValueDictionary

from gevent.event import AsyncResult
import sys
from datetime import datetime

__all__ = ["counter", "ResultsDictionary"]


class AsyncResult(AsyncResult):
    __slots__ = AsyncResult.__slots__ + ("ident", )


def counter(start=None, minimum=0, maximum=sys.maxsize - 1):
    # count = random.randint(minimum, maximum) if start is None else start
    count = int(datetime.now().timestamp()) if start is None else start
    while True:
        yield count
        count += 1
        if count >= maximum:
            count = int(datetime.now().timestamp())


class ResultsDictionary(WeakValueDictionary):

    def __init__(self):
        WeakValueDictionary.__init__(self)
        self._counter = counter()

    def pop(self, key, *args):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.pop(self, key, *args)

    def __contains__(self, key):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.__contains__(self, key)

    def __getitem__(self, key):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.__getitem__(self, key)

    def get(self, key, default=None):
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        return WeakValueDictionary.get(self, key, default=default)

    def __next__(self):
        result = AsyncResult()
        result.ident = ident = "%f.%f" % (next(self._counter), hash(result))
        self[ident] = result
        return result



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""The volttron.utils package contains generic utilities for handling json, storing configurations math
libraries...and more. """

import logging
from pathlib import Path
from typing import List

import yaml

from volttron.utils.commands import (is_volttron_running, execute_command, isapipe,
                                     wait_for_volttron_startup, wait_for_volttron_shutdown,
                                     vip_main)
from volttron.utils.context import ClientContext
from volttron.utils.commands import wait_for_volttron_startup, wait_for_volttron_shutdown
from volttron.utils.dynamic_helper import get_module, get_class, get_subclasses
from volttron.utils.file_access import create_file_if_missing
from volttron.utils.frame_serialization import serialize_frames, deserialize_frames
from volttron.utils.keystore import encode_key, decode_key
from volttron.utils.identities import normalize_identity, is_valid_identity
from volttron.utils.jsonapi import strip_comments, parse_json_config
from volttron.utils.logs import setup_logging, log_to_file
from volttron.utils.messagebus import store_message_bus_config
from volttron.utils.network import get_address, get_hostname, is_ip_private
from volttron.utils.time import (format_timestamp, process_timestamp, parse_timestamp_string,
                                 get_utc_seconds_from_epoch, get_aware_utc_now,
                                 fix_sqlite3_datetime)
from volttron.utils.version import get_version

_log = logging.getLogger(__name__)


def load_config(config_path):
    """Load a JSON-encoded configuration file."""
    if not config_path or not Path(config_path).exists():
        raise ValueError("Invalid config_path sent to function.")

    # First attempt parsing the file with a yaml parser (allows comments natively)
    # Then if that fails we fallback to our modified json parser.
    try:
        with open(config_path) as f:
            return yaml.safe_load(f.read())
    except yaml.YAMLError as e:
        try:
            with open(config_path) as f:
                return parse_json_config(f.read())
        except Exception as e:
            _log.error("Problem parsing agent configuration")
            raise


def update_kwargs_with_config(kwargs, config):
    """
    Loads the user defined configurations into kwargs and converts any dash/hyphen in config variables into underscores
    :param kwargs: kwargs to be updated
    :param config: dictionary of user/agent configuration
    """

    for k, v in config.items():
        kwargs[k.replace("-", "_")] = v


__all__: List[str] = [
    "update_kwargs_with_config", "load_config", "parse_json_config", "get_hostname", "log_to_file",
    "strip_comments", "setup_logging", "serialize_frames", "is_valid_identity", "isapipe",
    "is_volttron_running", "create_file_if_missing", "wait_for_volttron_shutdown",
    "process_timestamp", "parse_timestamp_string", "execute_command", "get_version",
    "get_aware_utc_now", "get_utc_seconds_from_epoch", "get_address", "deserialize_frames",
    "wait_for_volttron_startup", "normalize_identity", "ClientContext", "format_timestamp",
    "store_message_bus_config", "is_ip_private", "fix_sqlite3_datetime", "vip_main", "get_module",
    "get_class", "get_subclasses"
]



from __future__ import annotations

import inspect
import os
from dataclasses import dataclass
import logging
from os import PathLike
from pathlib import Path
from typing import Dict, List, Optional, Any, Type

import yaml

from volttron.utils.dynamic_helper import get_subclasses, get_class

_log = logging.getLogger(__name__)

def __all_ready_set__(property_name: str, value: Any):
    if value is not None:
        raise f"{property_name} has already been set and cannot be changed."


def __not_set__(property_name: str, value: Any):
    if not value:
        raise ValueError(f"{property_name} has not been set yet!")


@dataclass
class ServiceConfigs:
    """The ServiceConfigs object will manage the loading/storing of service configurations.

    Services will be loaded based upon the package they belong to.  By default services shall
    be loaded directly

    """
    service_config_file: Path
    server_config: ServerConfig

    def __post_init__(self):
        from volttron.server.serviceloader import discover_services
        self._loaded = {}
        if self.service_config_file.exists():
            self._loaded = yaml.safe_load(self.service_config_file.read_text().replace("$volttron_home",
                                                                                       os.environ['VOLTTRON_HOME']))

        self._namespace = self._loaded.get('namespace', 'volttron.services')
        self._discovered_services = discover_services(self._namespace)
        self._plugin_map = {}
        self._config_map = {}
        self._identity_map = {}
        self._instances = {}
        service_interface_cls = get_class("volttron.types", "ServiceInterface")

        for mod_name in self._discovered_services:
            try:
                cls = get_subclasses(mod_name, service_interface_cls)[0]
                identity = mod_name.replace("volttron.services", "platform")
                self._identity_map[mod_name] = identity
                self._plugin_map[mod_name] = cls
                self._config_map[mod_name] = self._loaded.get(mod_name, {})

            except ValueError:
                _log.warning(f"Couldn't load {mod_name}")
                continue

        # self.__auth_file__: Optional[PathLike] = None
        # self.__protected_topics_file__: Optional[PathLike] = None
        # self.__service_config_dict__: Dict = {}
        # self.__internal_address__: Optional[str] = None
        # self.__opts__ = None

    def init_services(self, server_config: ServerConfig):
        """
        Instantiate all the services available in the volttron.services namespace.
        :return:
        """

        for service_name, service_cls in self._plugin_map.items():
            # Look for parameter with server_config as an argument and pass that in if available
            # otherwise just pass kwargs
            params = inspect.signature(service_cls.__init__).parameters

            config = self._config_map.get(service_name)
            kwargs = {
                "identity": self._identity_map[service_name],
                "address": "inproc://vip"
            }
            for arg_name, arg_value in params.items():
                #arg_name, arg_value = arg
                if arg_name in config and arg_name != 'kwargs':
                    kwargs[arg_name] = config[arg_name]
                elif arg_name in config.get('kwargs', {}):
                    kwargs[arg_name] = config.get('kwargs')[arg_name]
                elif arg_name == 'aip':
                    kwargs[arg_name] = server_config.aip
                elif arg_name == 'server_config':
                    kwargs[arg_name] = server_config
            _log.info(f"Creating {service_name}")
            self._instances[service_name] = service_cls(**kwargs)

    def get_service_identity(self, service_name) -> str:
        return self._identity_map.get(service_name)

    def get_service_names(self) -> List[str]:
        return list(self._plugin_map.keys())

    def get_service(self, name: str) -> Type[ServiceInterface]:
        return self._plugin_map.get(name, None)

    def get_service_instance(self, name: str) -> ServiceInterface:
        return self._instances.get(name, None)

    def store_service_configs(self):
        pass
        # with self.service_config_file.open("wt") as fp:
        #     yaml.dump(self.)


class ServerConfig:

    def __init__(self):

        self.__aip__ = None
        self.__service_config_file__: Optional[PathLike] = None
        self.__auth_file__: Optional[PathLike] = None
        self.__protected_topics_file__: Optional[PathLike] = None
        self.__service_config_dict__: Dict = {}
        self.__internal_address__: Optional[str] = None
        self.__opts__ = None

    @property
    def opts(self):
        __not_set__("opts", self.__opts__)
        return self.__opts__

    @opts.setter
    def opts(self, value):
        __all_ready_set__("opts", self.__opts__)
        self.__opts__ = value

    @property
    def internal_address(self) -> str:
        __not_set__("internal_address", self.__internal_address__)
        return self.__internal_address__

    @internal_address.setter
    def internal_address(self, value):
        __all_ready_set__("internal_address", self.__internal_address__)
        self.__internal_address__ = value

    @property
    def service_config_file(self):
        __not_set__("service_config_file", self.__service_config_file__)
        return self.__service_config_file__

    @service_config_file.setter
    def service_config_file(self, value: PathLike):
        __all_ready_set__("service_config_file", self.__service_config_file__)
        if isinstance(value, str):
            value = Path(value)
        if not value.exists():
            raise ValueError(f"File {value} does not exists.")
        self.__service_config_file__ = value

    @property
    def aip(self) -> aipModule.AIPplatform:
        __not_set__("aip", self.__aip__)
        return self.__aip__

    @aip.setter
    def aip(self, value):
        if self.__aip__ is not None:
            raise ValueError("AIP has already been set and cannot be changed")
        self.__aip__ = value

    def get_service_enabled(self, service_name: str) -> bool:
        self.__init_service_dict__()
        service = self.__service_config_dict__.get(service_name)
        # Services don't have to be listed in the service_config.yml file so only if there is a
        # service listed do we consider whether they are enabled or not.
        retval = True
        if service is not None and "enabled" in service:
            retval = service["enabled"]
        return retval

    def get_service_kwargs(self, service_name: str) -> Dict:
        self.__init_service_dict__()
        service = self.__service_config_dict__.get(service_name, {})
        return {} if service is None else service.get("kwargs", {})

    def __init_service_dict__(self):
        if not self.__service_config_dict__:
            self.__service_config_dict__ = yaml.safe_load(self.__service_config_file__.open())

    @property
    def protected_topics_file(self) -> PathLike:
        return self.__protected_topics_file__

    @protected_topics_file.setter
    def protected_topics_file(self, value: PathLike):
        __all_ready_set__("protected_topics_file", self.__protected_topics_file__)
        if isinstance(value, str):
            value = Path(value)
        self.__protected_topics_file__ = value

    @property
    def auth_file(self) -> PathLike:
        __not_set__("auth_file", self.__auth_file__)
        return self.__auth_file__

    @auth_file.setter
    def auth_file(self, value: PathLike):
        __all_ready_set__("auth_file", self.__auth_file__)
        if isinstance(value, str):
            value = Path(value)
        self.__auth_file__ = value



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import os
import zmq
import logging
from volttron.utils import jsonapi
from zmq import EHOSTUNREACH, ZMQError, EAGAIN, NOBLOCK
from volttron.utils.frame_serialization import serialize_frames

_log = logging.getLogger(__name__)
# Optimizing by pre-creating frames
_ROUTE_ERRORS = {
    errnum: (
        zmq.Frame(str(errnum).encode("ascii")),
        zmq.Frame(os.strerror(errnum).encode("ascii")),
    )
    for errnum in [zmq.EHOSTUNREACH, zmq.EAGAIN]
}


class ExternalRPCService:
    """
    Class to manage routing of RPC calls between external platforms and internal agents(peers).
    """

    def __init__(self, socket, routing_service, *args, **kwargs):
        self._ext_router = routing_service
        self._vip_sock = socket
        # _log.debug("ExternalRPCService")

    def handle_subsystem(self, frames):
        """
         EXT_RPC subsystem handler on the server end.
        :frames list of frames
        :type frames list
        """
        response = []
        result = None

        try:
            sender, recipient, proto, usr_id, msg_id, subsystem, op, msg = frames[:9]
        except IndexError:
            return False

        if subsystem == "external_rpc":
            # If operation is to send to external platform
            if op == "send_platform":
                result = self._send_to_platform(frames)
            # If operation is to send to internal peer, use the internal router socket to send the frames
            elif op == "send_peer":
                result = self._send_to_peer(frames)
            if not result:
                response = result
            elif result is not None:
                # Form response frame
                response = [sender, recipient, proto, usr_id, msg_id, subsystem]
                response.append("request_response")
                response.append(result)
        return response

    def _send_to_platform(self, frames):
        """
        Send frames to external platform
        :param frames: frames following VIP format
        :return:
        """
        try:
            # Extract the frames and reorganize to add external platform and peer information
            sender, recipient, proto, usr_id, msg_id, subsystem, op, msg = frames[:9]
            # msg_data = jsonapi.loads(msg)
            msg_data = msg
            to_platform = msg_data["to_platform"]

            msg_data["from_platform"] = self._ext_router.my_instance_name()
            msg_data["from_peer"] = sender
            msg = jsonapi.dumps(msg_data)
            op = "send_peer"

            frames = ["", proto, usr_id, msg_id, subsystem, op, msg]
            # _log.debug("ROUTER: Sending EXT RernalPC message to: {}".format(to_platform))
            # Use external socket to send the message
            self._ext_router.send_external(to_platform, frames)
            return False
        except KeyError as exc:
            _log.error("Missing instance name in external RPC message: {}".format(exc))
        except IndexError:
            _log.error("Invalid EXT RPC message")

    def _send_to_peer(self, frames):
        """
        Send the external RPC message to local peer
        :param frames: Frames following VIP format
        :return: frames list
        """
        try:
            # Extract the frames and reorganize to send to local peer
            sender, recipient, proto, usr_id, msg_id, subsystem, op, msg = frames[:9]
            # msg_data = jsonapi.loads(msg)
            msg_data = msg
            peer = msg_data["to_peer"]
            frames[0] = peer
            drop = self._send_internal(frames)
            return False
        except KeyError as exc:
            _log.error("Missing agent name in external RPC message: {}".format(exc))
        except IndexError:
            _log.error("Invalid EXT RPC message")

    def _send_internal(self, frames):
        """
        Send message to internal/local peer
        :param frames: frames
        :return: peer to be dropped if not reachable
        """
        drop = []
        peer = frames[0]
        # Expecting outgoing frames:
        #   [RECIPIENT, SENDER, PROTO, USER_ID, MSG_ID, SUBSYS, ...]
        frames = serialize_frames(frames)
        try:
            # Try sending the message to its recipient
            self._vip_sock.send_multipart(frames, flags=NOBLOCK, copy=False)
        except ZMQError as exc:
            try:
                errnum, errmsg = error = _ROUTE_ERRORS[exc.errno]
            except KeyError:
                error = None
            if exc.errno == EHOSTUNREACH:
                _log.debug("Host unreachable {}".format(peer))
                drop.append(peer)
            elif exc.errno == EAGAIN:
                _log.debug("EAGAIN error {}".format(peer))
        return drop



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import inspect
import logging
import os
import sys
import traceback
import weakref
import re

import gevent.local
from gevent.event import AsyncResult
from volttron.utils import jsonapi

from .base import SubsystemBase
from ..results import counter, ResultsDictionary
from ..decorators import annotate, annotations, dualmethod, spawn
from volttron.utils import jsonrpc

from zmq import ZMQError
from zmq.green import ENOTSOCK

__all__ = ["RPC"]

_ROOT_PACKAGE_PATH = (os.path.dirname(__import__(__name__.split(".", 1)[0]).__path__[-1]) + os.sep)

_log = logging.getLogger(__name__)


def _isregex(obj):
    return (obj is not None and isinstance(obj, str) and len(obj) > 1 and obj[0] == obj[-1] == "/")


class Dispatcher(jsonrpc.Dispatcher):

    def __init__(self, methods, local):
        super(Dispatcher, self).__init__()
        self.methods = methods
        self.local = local
        self._results = ResultsDictionary()

    def serialize(self, json_obj):
        return jsonapi.dumps(json_obj)

    def deserialize(self, json_string):
        return jsonapi.loads(json_string)

    def batch_call(self, requests):
        methods = []
        results = []
        for notify, method, args, kwargs in requests:
            if notify:
                ident = None
            else:
                result = next(self._results)
                ident = result.ident
                results.append(result)
            methods.append((ident, method, args, kwargs))
        return super(Dispatcher, self).batch_call(methods), results

    def call(self, method, args=None, kwargs=None):
        # pylint: disable=arguments-differ
        result = next(self._results)
        return super(Dispatcher, self).call(result.ident, method, args, kwargs), result

    def result(self, response, ident, value, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set(value)

    def error(self, response, ident, code, message, data=None, context=None):
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(jsonrpc.exception_from_json(code, message, data))

    def exception(self, response, ident, message, context=None):
        # XXX: Should probably wrap exception in RPC specific error
        #      rather than re-raising.
        (
            _,    # exc_type
            exc,
            _,    # exc_tb
        ) = sys.exc_info()    # pylint: disable=unused-variable
        try:
            result = self._results.pop(ident)
        except KeyError:
            return
        result.set_exception(exc)

    def method(self, request, ident, name, args, kwargs, batch=None, context=None):
        if kwargs:
            try:
                args, kwargs = kwargs["*args"], kwargs["**kwargs"]
            except KeyError:
                pass
        try:
            method = self.methods[name]
        except KeyError:
            if name == "inspect":
                return {"methods": list(self.methods)}
            elif name.endswith(".inspect"):
                try:
                    method = self.methods[name[:-8]]
                except KeyError:
                    pass
                else:
                    return self._inspect(method)
            raise NotImplementedError(name)
        local = self.local
        local.vip_message = context
        local.request = request
        local.batch = batch
        try:
            return method(*args, **kwargs)
        except Exception as exc:    # pylint: disable=broad-except
            exc_tb = traceback.format_exc()
            _log.error("unhandled exception in JSON-RPC method %r: \n%s", name, exc_tb)
            if getattr(method, "traceback", True):
                exc.exc_info = {"exc_tb": exc_tb}
            raise
        finally:
            del local.vip_message
            del local.request
            del local.batch

    @staticmethod
    def _inspect(method):
        response = {"params": {}}
        signature = inspect.signature(method)
        for p in signature.parameters.values():
            response["params"][p.name] = {"kind": p.kind.name}
            if p.default is not inspect.Parameter.empty:
                response["params"][p.name]["default"] = p.default
            if p.annotation is not inspect.Parameter.empty:
                annotation = (p.annotation.__name__
                              if type(p.annotation) is type else str(p.annotation))
                response["params"][p.name]["annotation"] = annotation
        doc = inspect.getdoc(method)
        if doc:
            response["doc"] = doc
        try:
            source = inspect.getsourcefile(method)
            cut = len(os.path.commonprefix([_ROOT_PACKAGE_PATH, source]))
            source = source[cut:]
            lineno = inspect.getsourcelines(method)[1]
        except Exception:
            pass
        else:
            response["source"] = {"file": source, "line_number": lineno}
        ret = signature.return_annotation
        if ret is not inspect.Signature.empty:
            response["return"] = \
                ret.__name__ if type(ret) is type else str(ret)
        return response


class RPC(SubsystemBase):

    def __init__(self, core, owner, peerlist_subsys):
        self.core = weakref.ref(core)
        self._owner = owner
        self.context = None
        self._exports = {}
        self._dispatcher = None
        self._counter = counter()
        self._outstanding = weakref.WeakValueDictionary()
        core.register("RPC", self._handle_subsystem, self._handle_error)
        core.register(
            "external_rpc",
            self._handle_external_rpc_subsystem,
            self._handle_error,
        )
        self._isconnected = True
        self._message_bus = self.core().messagebus
        self.peerlist_subsystem = peerlist_subsys
        self.peer_list = {}

        def export(member):    # pylint: disable=redefined-outer-name
            for name in annotations(member, set, "rpc.exports"):
                self._exports[name] = member

        inspect.getmembers(owner, export)

        def setup(sender, **kwargs):
            # pylint: disable=unused-argument
            self.context = gevent.local.local()
            self._dispatcher = Dispatcher(self._exports, self.context)

        core.onsetup.connect(setup, self)
        core.ondisconnected.connect(self._disconnected)
        core.onconnected.connect(self._connected)
        self._iterate_exports()

    def _connected(self, sender, **kwargs):
        self._isconnected = True
        # Registering to 'onadd' and 'ondrop' signals to get notified
        # whenever new peer is added/removed
        self.peerlist_subsystem.onadd.connect(self._add_new_peer)
        self.peerlist_subsystem.ondrop.connect(self._drop_new_peer)

    def _disconnected(self, sender, **kwargs):
        self._isconnected = False

    def _add_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            message_bus = kwargs.pop("message_bus")
            self.peer_list[peer] = message_bus
        except KeyError:
            pass

    def _drop_new_peer(self, sender, **kwargs):
        try:
            peer = kwargs.pop("peer")
            self.peer_list.pop(peer)
        except KeyError:
            pass

    def _iterate_exports(self):
        """
        Iterates over exported methods and adds authorization checks
        as necessary
        """
        for method_name in self._exports:
            method = self._exports[method_name]
            caps = annotations(method, set, "rpc.allow_capabilities")
            if caps:
                self._exports[method_name] = self._add_auth_check(method, caps)

    def _add_auth_check(self, method, required_caps):
        """
        Adds an authorization check to verify the calling agent has the
        required capabilities.
        """

        def checked_method(*args, **kwargs):
            user = str(self.context.vip_message.user)
            if self._message_bus == "rmq":
                # When we address issue #2107 external platform user should
                # have instance name also included in username.
                user = user.split(".")[1]
            user_capabilites = self._owner.vip.auth.get_capabilities(user)
            _log.debug("**user caps is: {}".format(user_capabilites))
            if user_capabilites:
                user_capabilities_names = set(user_capabilites.keys())
            else:
                user_capabilities_names = set()
            if required_caps == {""}:
                pass
            elif not required_caps.issubset(user_capabilities_names):
                msg = ("method '{}' requires capabilities {}, but capability {} "
                       "was provided for user {}").format(method.__name__, required_caps,
                                                          user_capabilites, user)
                raise jsonrpc.exception_from_json(jsonrpc.UNAUTHORIZED, msg)
            else:
                # Now check if args passed to method are the ones allowed.

                for cap_name, param_dict in user_capabilites.items():
                    if (param_dict and required_caps and cap_name in required_caps):
                        # if the method has required capabilities and
                        # if the user capability has argument restrictions,
                        # check if the args passed to method match the
                        # requirement
                        _log.debug("args = {} kwargs= {}".format(args, kwargs))
                        args_dict = inspect.getcallargs(method, *args, **kwargs)
                        _log.debug("dict = {}".format(args_dict))
                        _log.debug("name= %r parameters allowed=%r", cap_name, param_dict)
                        for name, value in param_dict.items():
                            _log.debug("name= {} value={}".format(name, value))
                            if name not in args_dict:
                                raise jsonrpc.exception_from_json(
                                    jsonrpc.UNAUTHORIZED,
                                    "User {} capability is not defined "
                                    "properly. method {} does not have "
                                    "a parameter {}".format(user, method.__name__, name),
                                )
                            if _isregex(value):
                                regex = re.compile("^" + value[1:-1] + "$")
                                if not regex.match(args_dict[name]):
                                    raise jsonrpc.exception_from_json(
                                        jsonrpc.UNAUTHORIZED,
                                        "User {} can call method {} only "
                                        "with {} matching pattern {} but "
                                        "called with {}={}".format(
                                            user,
                                            method.__name__,
                                            name,
                                            value,
                                            name,
                                            args_dict[name],
                                        ),
                                    )
                            elif args_dict[name] != value:
                                raise jsonrpc.exception_from_json(
                                    jsonrpc.UNAUTHORIZED,
                                    "User {} can call method {} only "
                                    "with {}={} but called with "
                                    "{}={}".format(
                                        user,
                                        method.__name__,
                                        name,
                                        value,
                                        name,
                                        args_dict[name],
                                    ),
                                )

            return method(*args, **kwargs)

        return checked_method

    @spawn
    def _handle_external_rpc_subsystem(self, message):
        ret_msg = dict()
        operation = message.args[0]
        rpc_msg = message.args[1]    # jsonapi.loads(message.args[1])
        try:
            method_args = rpc_msg["args"]
            # message.args = [method_args]
            message.args = method_args
            for idx, msg in enumerate(message.args):
                if isinstance(msg, str):
                    message.args[idx] = jsonapi.loads(msg)
            dispatch = self._dispatcher.dispatch
            # _log.debug("External RPC IN message args {}".format(message))

            responses = [
                response for response in (dispatch(msg, message) for msg in message.args)
                if response
            ]
            # _log.debug("External RPC Responses {}".format(responses))
            if responses:
                message.user = ""
                try:
                    message.peer = ""
                    message.subsystem = "external_rpc"
                    frames = []
                    operation = "send_platform"
                    frames.append(operation)
                    msg = jsonapi.dumps(
                        dict(
                            to_platform=rpc_msg["from_platform"],
                            to_peer=rpc_msg["from_peer"],
                            from_platform=rpc_msg["to_platform"],
                            from_peer=rpc_msg["to_peer"],
                            args=responses,
                        ))
                    frames.append(msg)
                except KeyError:
                    _log.error("External RPC message did not contain "
                               "proper message format")
                message.args = jsonapi.dumps(ret_msg)
                try:
                    self.core().connection.send_vip(
                        peer="",
                        subsystem="external_rpc",
                        args=frames,
                        msg_id=message.id,
                        user=message.user,
                        copy=False,
                    )
                except ZMQError as ex:
                    _log.error("ZMQ error: {}".format(ex))
                    pass
        except KeyError:
            pass

    @spawn
    def _handle_subsystem(self, message):
        dispatch = self._dispatcher.dispatch

        if self._message_bus == "rmq":
            for idx, msg in enumerate(message.args):
                if not isinstance(msg, dict):
                    message.args[idx] = jsonapi.loads(msg)

            responses = [
                response for response in (dispatch(msg, message) for msg in message.args)
                if response
            ]
        else:
            responses = [
                response for response in (dispatch(msg, message) for msg in message.args)
                if response
            ]
        if responses:
            message.user = ""
            message.args = responses
            try:
                if self._isconnected:
                    if self._message_bus == "zmq":
                        self.core().connection.send_vip_object(message, copy=False)
                    else:
                        # Agent is running on RMQ message bus.
                        # Adding backward compatibility support for ZMQ.
                        # Check if the peer is running on ZMQ bus.
                        # If yes, send RPC message to proxy router
                        # agent to forward using ZMQ message bus connection
                        try:
                            msg_bus = self.peer_list[message.peer]
                        except KeyError:
                            msg_bus = self._message_bus
                        if msg_bus == "zmq":
                            # If peer connected to ZMQ bus,
                            # send via proxy router agent
                            self.core().connection.send_vip_object_via_proxy(message)
                        else:
                            self.core().connection.send_vip_object(message, copy=False)
            except ZMQError as exc:
                if exc.errno == ENOTSOCK:
                    _log.debug("Socket send on non-socket %s", self.core().identity)

    def _handle_error(self, sender, message, error, **kwargs):
        result = self._outstanding.pop(message.id, None)
        if isinstance(result, AsyncResult):
            result.set_exception(error)
        elif result:
            for result in result:
                result.set_exception(error)

    def get_exports(self):
        """Returns a list copy of all exported methods."""
        return [method for method in self._exports].copy()

    @dualmethod
    def export(self, method, name=None):
        self._exports[name or method.__name__] = method
        return method

    @export.classmethod
    def export(cls, name=None):    # pylint: disable=no-self-argument
        if name is not None and not isinstance(name, str):
            method, name = name, name.__name__
            annotate(method, set, "rpc.exports", name)
            return method

        def decorate(method):
            annotate(method, set, "rpc.exports", name)
            return method

        return decorate

    def batch(self, peer, requests):
        request, results = self._dispatcher.batch_call(requests)
        if results:
            items = weakref.WeakSet(results)
            ident = "%s.%s" % (next(self._counter), id(items))
            for result in results:
                result._weak_set = items    # pylint: disable=protected-access
            self._outstanding[ident] = items
        else:
            ident = ""
        if request:
            if self._isconnected:
                try:
                    self.core().connection.send_vip(peer, "RPC", [request], msg_id=ident)
                except ZMQError as exc:
                    if exc.errno == ENOTSOCK:
                        _log.debug("Socket send on non-socket %r", self.core().identity)
        return results or None

    def call(self, peer, method, *args, **kwargs):
        platform = kwargs.pop("external_platform", "")
        request, result = self._dispatcher.call(method, args, kwargs)
        ident = f"{next(self._counter)}.{hash(result)}"
        self._outstanding[ident] = result
        subsystem = None
        frames = []

        if not self._isconnected:
            return

        if self._message_bus == "zmq":
            if platform == "":    # local platform
                subsystem = "RPC"
                frames.append(request)
            else:
                frames = []
                operation = "send_platform"
                subsystem = "external_rpc"
                frames.append(operation)
                msg = dict(
                    to_platform=platform,
                    to_peer=peer,
                    from_platform="",
                    from_peer="",
                    args=[request],
                )
                frames.append(msg)
                peer = ""

            try:
                self.core().connection.send_vip(peer, subsystem, args=frames, msg_id=ident)
            except ZMQError as exc:
                if exc.errno == ENOTSOCK:
                    _log.debug("Socket send on non-socket %r", self.core().identity)
        else:
            # Agent running on RMQ message bus.
            # Adding backward compatibility support for ZMQ. Check if peer
            # is running on ZMQ bus. If yes, send RPC message to proxy router
            # agent to forward over ZMQ message bus connection
            try:
                peer_msg_bus = self.peer_list[peer]
            except KeyError:
                peer_msg_bus = self._message_bus
            if peer_msg_bus == "zmq":
                # peer connected to ZMQ bus, send via proxy router agent
                self.core().connection.send_via_proxy(peer, "RPC", msg_id=ident, args=[request])
            else:
                self.core().connection.send_vip(peer,
                                                "RPC",
                                                args=[request],
                                                msg_id=ident,
                                                platform=platform)

        return result

    __call__ = call

    def notify(self, peer, method, *args, **kwargs):
        platform = kwargs.pop("external_platform", "")
        request = self._dispatcher.notify(method, args, kwargs)
        frames = []
        if not self._isconnected:
            return

        if self._message_bus == "zmq":
            subsystem = None
            if platform == "":
                subsystem = "RPC"
                frames.append(request)
            else:
                operation = "send_platform"
                subsystem = "external_rpc"
                frames.append(operation)
                msg = dict(
                    to_platform=platform,
                    to_peer=peer,
                    from_platform="",
                    from_peer="",
                    args=[request],
                )
                frames.append(msg)
                peer = ""

            try:
                self.core().connection.send_vip(peer, subsystem, args=frames)
            except ZMQError as exc:
                if exc.errno == ENOTSOCK:
                    _log.debug("Socket send on non-socket %r", self.core().identity)
        else:
            # Agent running on RMQ message bus.
            # Adding backward compatibility support for ZMQ. Check if peer
            # is running on ZMQ bus. If yes, send RPC message to proxy
            # router agent to forward over ZMQ message bus connection
            try:
                peer_msg_bus = self.peer_list[peer]
            except KeyError:
                peer_msg_bus = self._message_bus
            if peer_msg_bus == "zmq":
                # peer connected to ZMQ bus, send via proxy router agent
                self.core().connection.send_via_proxy(peer, "RPC", args=[request])
            else:
                self.core().connection.send_vip(peer, "RPC", args=[request], platform=platform)

    @dualmethod
    def allow(self, method, capabilities):
        if isinstance(capabilities, str):
            cap = set([capabilities])
        else:
            cap = set(capabilities)
        # Necessary if you have provided an alias for the rpc method.
        if isinstance(method, str):
            if method in self._exports:
                self._exports[method] = self._add_auth_check(self._exports[method], cap)
            else:
                _log.error("Method alias is not in RPC export list.")
        else:
            self._exports[method.__name__] = self._add_auth_check(method, cap)

    @allow.classmethod
    def allow(cls, capabilities):
        """
        Decorator specifies required agent capabilities to call a method.

        This is designed to be used with the export decorator:

        .. code-block:: python

            @RPC.export
            @RPC.allow('can_read_status')
            def get_status():
                ...

        Multiple capabilities can be provided in a list:
        .. code-block:: python

            @RPC.allow(['can_read_status', 'can_call_my_methods'])

        """

        def decorate(method):
            if isinstance(capabilities, str):
                annotate(method, set, "rpc.allow_capabilities", capabilities)
            else:
                for cap in capabilities:
                    annotate(method, set, "rpc.allow_capabilities", cap)
            return method

        return decorate



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

AUTH = "platform.auth"

VOLTTRON_CENTRAL = "volttron.central"
VOLTTRON_CENTRAL_PLATFORM = "platform.agent"

PLATFORM_DRIVER = "platform.driver"
PLATFORM_TOPIC_WATCHER = "platform.topic_watcher"
PLATFORM_SYSMON = "platform.sysmon"
PLATFORM_EMAILER = "platform.emailer"
PLATFORM_HEALTH = "platform.health"
# The PLATFORM_ALERTER known name is now deprecated
PLATFORM_ALERTER = PLATFORM_TOPIC_WATCHER
PLATFORM_HISTORIAN = "platform.historian"

PLATFORM_MARKET_SERVICE = "platform.market"

ROUTER = ""
CONTROL = "platform.control"
CONTROL_CONNECTION = "control.connection"
PLATFORM_WEB = "platform.web"
CONFIGURATION_STORE = "platform.config_store"
KEY_DISCOVERY = "keydiscovery"
PROXY_ROUTER = "zmq.proxy.router"

ALL_KNOWN_IDENTITIES = sorted((
    ROUTER,
    VOLTTRON_CENTRAL,
    VOLTTRON_CENTRAL_PLATFORM,
    PLATFORM_HISTORIAN,
    CONTROL,
    CONTROL_CONNECTION,
    PLATFORM_WEB,
    AUTH,
    PLATFORM_TOPIC_WATCHER,
    CONFIGURATION_STORE,
    PLATFORM_MARKET_SERVICE,
    PLATFORM_EMAILER,
    PLATFORM_SYSMON,
    PLATFORM_HEALTH,
    KEY_DISCOVERY,
    PROXY_ROUTER,
))

PROCESS_IDENTITIES = sorted((
    AUTH,
    PLATFORM_HEALTH,
    CONFIGURATION_STORE,
    CONTROL,
    PLATFORM_WEB,
    KEY_DISCOVERY,
    PROXY_ROUTER,
))



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""Dumping ground for VOLTTRON platform™ agent math helper functions.

Not meant to replace numpy in all cases. A basic set common math
routines to remove the need for numpy in simple cases.

This module should NEVER import numpy as that would defeat the
purpose.
"""

from math import sqrt


def mean(data):
    """Return the sample arithmetic mean of data."""
    n = len(data)
    if n < 1:
        raise ValueError('mean requires at least one data point')
    return sum(data) / n    # in Python 2 use sum(data)/float(n)


def _ss(data):
    """Return sum of square deviations of sequence data."""
    c = mean(data)
    ss = sum((x - c)**2 for x in data)
    return ss


def pstdev(data):
    """Calculate the population standard deviation."""
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss / n    # the population variance
    return sqrt(pvar)


def stdev(data):
    """Calculate the sample standard deviation."""
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss / (n - 1)    # sample variance
    return sqrt(pvar)



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""VIP - VOLTTRON™ Interconnect Protocol implementation

See https://volttron.readthedocs.io/en/develop/core_services/messagebus/VIP/VIP-Overview.html
for protocol specification.

This module is for use within gevent. It provides some locking around
send operations to protect the VIP state. It should be safe to use a
single socket in multiple greenlets without any kind of locking.
"""

from contextlib import contextmanager as _contextmanager

from gevent import sleep as _sleep
from gevent.local import local as _local
from gevent.lock import RLock as _RLock

from zmq.green import NOBLOCK, POLLOUT
from zmq import green as _green

from volttron.utils.socket import _Socket
# from volttron.client.vip.router import BaseRouter as _BaseRouter


class Socket(_Socket, _green.Socket):
    _context_class = _green.Context
    _local_class = _local

    def __init__(self, *args, **kwargs):
        super(Socket, self).__init__(*args, **kwargs)
        object.__setattr__(self, "_Socket__send_lock", _RLock())

    @_contextmanager
    def _sending(self, flags):
        flags |= getattr(self._Socket__local, "flags", 0)
        lock = self._Socket__send_lock
        while not lock.acquire(not flags & NOBLOCK):
            if not self.poll(0, POLLOUT):
                raise _green.Again()
            _sleep(0)
        try:
            yield flags
        finally:
            lock.release()


# class BaseRouter(_BaseRouter):
#     _context_class = _green.Context
#     _socket_class = _green.Socket



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufactufrer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import zmq
import logging

from ..vip.green import Socket as GreenSocket
from . import BaseConnection

# TODO ADD BACK rmq
# from volttron.client.vip.rmq_connection import BaseConnection
_log = logging.getLogger(__name__)


class ZMQConnection(BaseConnection):
    """
    Maintains ZMQ socket connection
    """

    def __init__(self, url, identity, instance_name, context):
        super(ZMQConnection, self).__init__(url, identity, instance_name)

        self.socket = None
        self.context = context
        self._identity = identity
        self._logger = logging.getLogger(__name__)
        self._logger.debug("ZMQ connection {}".format(identity))

    def open_connection(self, type):
        if type == zmq.DEALER:
            self.socket = GreenSocket(self.context)
            if self._identity:
                self.socket.identity = self._identity.encode("utf-8")
        else:
            self.socket = zmq.Socket()

    def set_properties(self, flags):
        hwm = flags.get("hwm", 6000)
        self.socket.set_hwm(hwm)
        reconnect_interval = flags.get("reconnect_interval", None)
        if reconnect_interval:
            self.socket.setsockopt(zmq.RECONNECT_IVL, reconnect_interval)

    def connect(self, callback=None):
        _log.debug(f"connecting to url {self._url}")
        _log.debug(f"url type is {type(self._url)}")

        self.socket.connect(self._url)
        if callback:
            callback(True)

    def bind(self):
        pass

    def register(self, handler):
        self._vip_handler = handler

    def send_vip_object(self, message, flags=0, copy=True, track=False):
        self.socket.send_vip_object(message, flags, copy, track)

    def send_vip(
        self,
        peer,
        subsystem,
        args=None,
        msg_id: bytes = b"",
        user=b"",
        via=None,
        flags=0,
        copy=True,
        track=False,
    ):
        self.socket.send_vip(
            peer,
            subsystem,
            args=args,
            msg_id=msg_id,
            user=user,
            via=via,
            flags=flags,
            copy=copy,
            track=track,
        )

    def recv_vip_object(self, flags=0, copy=True, track=False):
        return self.socket.recv_vip_object(flags, copy, track)

    def disconnect(self):
        self.socket.disconnect(self._url)

    def close_connection(self, linger=5):
        """This method closes ZeroMQ socket"""
        self.socket.close(linger)
        _log.debug("********************************************************************")
        _log.debug("Closing connection to ZMQ: {}".format(self._identity))
        _log.debug("********************************************************************")



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

from collections import defaultdict
from datetime import datetime
import logging

from volttron.client.known_identities import CONTROL_CONNECTION, PROCESS_IDENTITIES
from volttron.types import ServiceInterface
from volttron.utils import format_timestamp
from volttron.client.vip.agent import Agent, Core, RPC

# TODO: rmq addition
# from volttron.utils.rmq_config_params import RMQConfig
# from volttron.utils.rmq_setup import start_rabbit, RabbitMQStartError

_log = logging.getLogger(__name__)


class HealthService(ServiceInterface):

    def __init__(self, **kwargs):
        super(HealthService, self).__init__(**kwargs)

        # Store the health stats for given peers in a dictionary with
        # keys being the identity of the connected agent.
        self._health_dict = defaultdict(dict)

    def peer_added(self, peer):
        """
        The `peer_added` method should be called whenever an agent is connected to the
        platform.

        :param peer: The identity of the agent connected to the platform
        """
        health = self._health_dict[peer]

        health["peer"] = peer
        health["service_agent"] = peer in PROCESS_IDENTITIES
        health["connected"] = format_timestamp(datetime.now())

    def peer_dropped(self, peer):
        # TODO: Should there be an option for  a db/log file for agents coming and going from the platform?
        self._health_dict[peer]["disconnected"] = format_timestamp(datetime.now())
        del self._health_dict[peer]

    @RPC.export
    def get_platform_health(self):
        """
        The `get_platform_health` retrieves all of the connected agent's health structures,
        except for the `CONTROL_CONNECTION` (vctl's known identity).  Vctl's identity is used for short
        term connections and is not relevant to the core health system.

        This function returns a dictionary in the form identity: values such as the following:

        .. code-block :: json

            {
                "listeneragent-3.3_35":
                {
                    "peer": "listeneragent-3.3_35",
                    "service_agent": False,
                    "connected": "2020-10-28T12:46:58.701119",
                    "last_heartbeat": "2020-10-28T12:47:03.709605",
                    "message": "GOOD"
                }
            }

        :return:
        """
        # Ignore the connection from control in the health as it will only be around for a short while.
        agents = {
            k: v
            for k, v in self._health_dict.items() if not v.get("peer") == CONTROL_CONNECTION
        }
        return agents

    def _heartbeat_updates(self, peer, sender, bus, topic, headers, message):
        """
        This method is called whenever a publish goes on the message bus from the
        heartbeat* topic.

        :param peer:
        :param sender:
        :param bus:
        :param topic:
        :param headers:
        :param message:
        :return:
        """
        health = self._health_dict[sender]
        time_now = format_timestamp(datetime.now())
        if not health:
            health["connected"] = time_now
            health["peer"] = sender
            health["service_agent"] = sender in PROCESS_IDENTITIES

        health["last_heartbeat"] = time_now
        health["message"] = message

    @Core.receiver("onstart")
    def onstart(self, sender, **kwargs):
        # Start subscribing to heartbeat topic to get updates from the health subsystem.
        self.vip.pubsub.subscribe("pubsub", "heartbeat", callback=self._heartbeat_updates)



import logging
import os
import stat
import sys
import syslog
import traceback
import warnings

from volttron.utils import jsonapi

try:
    HAS_SYSLOG = True
    import syslog
except ImportError:
    HAS_SYSLOG = False

# Keep the ability to have system log output for linux
# this will fail on windows because no syslog.
if HAS_SYSLOG:

    class SyslogFormatter(logging.Formatter):
        _level_map = {
            logging.DEBUG: syslog.LOG_DEBUG,
            logging.INFO: syslog.LOG_INFO,
            logging.WARNING: syslog.LOG_WARNING,
            logging.ERROR: syslog.LOG_ERR,
            logging.CRITICAL: syslog.LOG_CRIT,
        }

        def format(self, record):
            level = self._level_map.get(record.levelno, syslog.LOG_INFO)
            return "<{}>".format(level) + super(SyslogFormatter, self).format(record)


def isapipe(fd):
    fd = getattr(fd, "fileno", lambda: fd)()
    return stat.S_ISFIFO(os.fstat(fd).st_mode)


class JsonFormatter(logging.Formatter):

    def format(self, record):
        dct = record.__dict__.copy()
        dct["msg"] = record.getMessage()
        dct.pop("args")
        exc_info = dct.pop("exc_info", None)
        if exc_info:
            dct["exc_text"] = "".join(traceback.format_exception(*exc_info))
        return jsonapi.dumps(dct)


class AgentFormatter(logging.Formatter):

    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = "%(asctime)s %(composite_name)s %(levelname)s: %(message)s"
        super(AgentFormatter, self).__init__(fmt=fmt, datefmt=datefmt)

    def composite_name(self, record):
        if record.name == "agents.log":
            cname = "(%(processName)s %(process)d) %(remote_name)s"
        elif record.name.startswith("agents.std"):
            cname = "(%(processName)s %(process)d) <{}>".format(record.name.split(".", 2)[1])
        else:
            cname = "() %(name)s"
        return cname % record.__dict__

    def format(self, record):
        if "composite_name" not in record.__dict__:
            record.__dict__["composite_name"] = self.composite_name(record)
        return super(AgentFormatter, self).format(record)


class FramesFormatter(object):

    def __init__(self, frames):
        self.frames = frames

    def __repr__(self):
        output = ''
        for f in self.frames:
            output += str(f)
        return output

    __str__ = __repr__


def log_to_file(file, level=logging.WARNING, handler_class=logging.StreamHandler):
    """Direct log output to a file (or something like one)."""
    handler = handler_class(file)
    handler.setLevel(level)
    handler.setFormatter(
        AgentFormatter("%(asctime)s %(composite_name)s %(levelname)s: %(message)s"))
    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(handler)


def setup_logging(level=logging.DEBUG, console=False):
    root = logging.getLogger()
    if not root.handlers:
        handler = logging.StreamHandler()

        if isapipe(sys.stderr) and "_LAUNCHED_BY_PLATFORM" in os.environ:
            handler.setFormatter(JsonFormatter())
        elif console:
            # Below format is more readable for console
            handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        else:
            fmt = "%(asctime)s %(name)s %(levelname)s: %(message)s"
            handler.setFormatter(logging.Formatter(fmt))
        if level != logging.DEBUG:
            # import it here so that when urllib3 imports the requests package, ssl would already got
            # monkey patched by gevent.
            # and this warning is needed only when log level is not debug
            from urllib3.exceptions import InsecureRequestWarning

            warnings.filterwarnings("ignore", category=InsecureRequestWarning)
        root.addHandler(handler)
    root.setLevel(level)



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging
import weakref

from .base import SubsystemBase
from ..results import ResultsDictionary
from zmq import ZMQError
from zmq.green import ENOTSOCK

__all__ = ["Ping"]

_log = logging.getLogger(__name__)


class Ping(SubsystemBase):

    def __init__(self, core):
        self.core = weakref.ref(core)
        self._results = ResultsDictionary()
        core.register("ping", self._handle_ping, self._handle_error)

    def ping(self, peer, *args):
        result = next(self._results)
        args = list(args)
        args.insert(0, "ping")
        connection = self.core().connection
        try:
            connection.send_vip("", "ping", args=["drop", peer], msg_id=result.ident)
        except ZMQError as exc:
            if exc.errno == ENOTSOCK:
                _log.debug("Socket send on non socket {}".format(self.core().identity))
        return result

    __call__ = ping

    def _handle_ping(self, message):
        connection = self.core().connection
        try:
            op = message.args[0]
        except IndexError:
            _log.error("missing ping subsystem operation")
            return
        if op == "ping":
            message.user = ""
            message.args[0] = "pong"
            connection.send_vip_object(message, copy=False)
        elif op == "pong":
            try:
                result = self._results.pop(message.id)
            except KeyError:
                return
            result.set([arg for arg in message.args[1:]])
        else:
            _log.error("unknown ping subsystem operation")

    def _handle_error(self, sender, message, error, **kwargs):
        try:
            result = self._results.pop(message.id)
        except KeyError:
            return
        result.set_exception(error)



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

from __future__ import print_function, absolute_import

import logging

from zmq import green as zmq
from zmq.green import ZMQError, ENOTSOCK

from volttron.utils import ClientContext as cc
from volttron.utils import jsonapi
from volttron.utils.frame_serialization import deserialize_frames, serialize_frames

from volttron.client.vip.agent import Agent, Core

_log = logging.getLogger(__name__)

if cc.is_rabbitmq_available():
    import pika


class ZMQProxyRouter(Agent):
    """
    Proxy ZMQ based router agent is implemented for backward compatibility with ZeroMQ based message bus. In a single
    instance setup, either ZeroMQ or RabbitMQ based message bus will be running and all the agents will be using
    the same message bus. But in multi-platform setup, some instances maybe running with RabbitMQ message bus and
    others with ZeroMQ message bus. The Proxy router agent is implemented to manage the routing between local and
    external instances in such cases.

    Please note, if all instances in multi-platform setup are RabbitMQ based, then RabbitMQ federation/shovel need to
    be used.
    """

    def __init__(self, address, identity, zmq_router, *args, **kwargs):
        super(ZMQProxyRouter, self).__init__(identity, address, **kwargs)
        self.zmq_router = zmq_router
        self._routing_key = self.core.instance_name + "." + "proxy"
        rmq_user = self.core.instance_name + "." + identity
        self._outbound_response_queue = "{user}.zmq.outbound.response".format(user=rmq_user)
        self._outbound_request_queue = "{user}.zmq.outbound.request".format(user=rmq_user)
        self._rpc_handler_queue = "{user}.zmq.outbound.subsystem".format(user=rmq_user)
        self._vip_loop_running = False
        self._zmq_peers = set()

    @Core.receiver("onstart")
    def startup(self, sender, **kwargs):
        """
        On startup, it does the following:
         - Start ZMQ Router loop.
         - Establish RMQ queue bindings to handle routing of messages
         between internal and external agents.
        :param sender:
        :param kwargs:
        :return:
        """
        if not self._vip_loop_running:
            self.core.spawn(self.vip_loop)

        connection = self.core.connection
        channel = connection.channel

        # ----------------------------------------------------------------------------------
        # Create a queue to receive messages from local platform
        # (for example, response for RPC request etc)
        result = channel.queue_declare(
            queue=self._rpc_handler_queue,
            durable=False,
            exclusive=True,
            auto_delete=True,
            callback=None,
        )
        channel.queue_bind(
            exchange=connection.exchange,
            queue=self._rpc_handler_queue,
            routing_key=self.core.instance_name + ".proxy.router.zmq.outbound.subsystem",
            callback=None,
        )
        channel.basic_consume(self.rpc_message_handler, queue=self._rpc_handler_queue, no_ack=True)
        # --------------------------------------------------------------------------------------

        # Create a queue to receive messages from local platform
        # (for example, response for RPC request etc)
        result = channel.queue_declare(
            queue=self._outbound_response_queue,
            durable=False,
            exclusive=True,
            auto_delete=True,
            callback=None,
        )
        channel.queue_bind(
            exchange=connection.exchange,
            queue=self._outbound_response_queue,
            routing_key=self.core.instance_name + ".proxy.router.subsystems",
            callback=None,
        )
        channel.basic_consume(
            self.outbound_response_handler,
            queue=self._outbound_response_queue,
            no_ack=True,
        )

        # Create a queue to receive messages from local platform.
        # For example, external platform pubsub/RPC subscribe/unsubscribe
        # requests from internal agents
        channel.queue_declare(
            queue=self._outbound_request_queue,
            durable=False,
            exclusive=True,
            auto_delete=True,
            callback=None,
        )
        # Binding for external platform pubsub message requests
        channel.queue_bind(
            exchange=connection.exchange,
            queue=self._outbound_request_queue,
            routing_key=self.core.instance_name + ".proxy.router.pubsub",
            callback=None,
        )

        # Binding for external platform RPC message requests
        channel.queue_bind(
            exchange=connection.exchange,
            queue=self._outbound_request_queue,
            routing_key=self.core.instance_name + ".proxy.router.external_rpc",
            callback=None,
        )
        channel.basic_consume(
            self.outbound_request_handler,
            queue=self._outbound_request_queue,
            no_ack=True,
        )

    @Core.receiver("onstop")
    def on_stop(self, sender, **kwargs):
        """
        Stop the ZMQ router
        :param sender:
        :param kwargs:
        :return:
        """
        _log.debug("********************************************************************")
        _log.debug("Stopping ZMQ Router")
        _log.debug("********************************************************************")
        self.zmq_router.stop()

    def outbound_response_handler(self, ch, method, props, body):
        """
        Message received from internal agent to send to remote agent in ZMQ VIP message format.
        :param ch: channel
        :param method: contains routing key
        :param props: message properties like VIP header information
        :param body: message
        :return:
        """
        # Strip sender's identity from binding key
        routing_key = str(method.routing_key)
        platform, to_identity = routing_key.split(".", 1)
        platform, from_identity = props.app_id.split(".", 1)
        userid = props.headers.get("user", "")
        # Reformat message into ZMQ VIP format
        frames = [
            to_identity,
            from_identity,
            "VIP1",
            userid,
            props.message_id,
            props.type,
        ]

        try:
            args = jsonapi.loads(body)
            try:
                # This is necessary because jsonrpc request/response is inside a list which the
                # ZMQ agent subsystem does not like
                args = jsonapi.loads(args[0])
                frames.append(jsonapi.dumps(args))
            except ValueError as e:
                if isinstance(args, list):
                    for m in args:
                        frames.append(m)
                else:
                    frames.append(jsonapi.dumps(args))
        except TypeError as e:
            _log.error("Invalid json format {}".format(e))
            return

        _log.debug("Proxy ZMQ Router Outbound handler {0}, {1}".format(to_identity, args))
        frames = serialize_frames(frames)
        try:
            self.zmq_router.socket.send_multipart(frames, copy=True)
        except ZMQError as ex:
            _log.error("ZMQ Error {}".format(ex))

    def rpc_message_handler(self, ch, method, props, body):
        """

        :param ch:
        :param method:
        :param props:
        :param body:
        :return:
        """

        frames = serialize_frames(jsonapi.loads(body))

        try:
            self.zmq_router.socket.send_multipart(frames, copy=False)
        except ZMQError as ex:
            _log.error("ZMQ Error {}".format(ex))

    def outbound_request_handler(self, ch, method, props, body):
        """
        Handler for receiving external platform PubSub/RPC requests from internal agents.
        It then calls external PubSub/RPC router handler to forward the request to external platform.
        :param ch: channel
        :param method: contains the routing key
        :param props: message properties
        :param body: message body
        :return:
        """
        _log.debug("Proxy ZMQ Router {}".format(body))
        frames = jsonapi.loads(body.decode("utf-8"))
        if len(frames) > 6:
            if frames[5] == "pubsub":
                # Forward the message to pubsub component of the router to take action
                self.zmq_router.pubsub.handle_subsystem(frames)
            else:
                # Forward the message to external RPC handler component of the router to take action
                self.zmq_router.ext_rpc.handle_subsystem(frames)

    def publish_callback(self, peer, sender, bus, topic, headers, message):
        """
        Callback method registered with local message bus to receive PubSub messages
        subscribed by external platform agents. PubSub component of router will route the message to
        appropriate external platform subscribers.
        :return:
        """
        json_msg = jsonapi.dumps(dict(bus=bus, headers=headers, message=message))
        # Reformat the message into ZMQ VIP message frames
        frames = [
            sender,
            "",
            "VIP",
            "",
            "",
            "pubsub",
            zmq.Frame("publish"),
            zmq.Frame(str(topic)),
            zmq.Frame(str(json_msg)),
        ]

        self.zmq_router.pubsub.handle_subsystem(frames, "")

    def vip_loop(self):
        """
        Infinite VIP loop to receive and send messages over ZMQ message bus.
        :return:
        """
        connection = self.core.connection
        # Set up ZMQ router socket connection
        self.zmq_router.start()
        # Register proxy agent handle
        self.zmq_router.pubsub.add_rabbitmq_agent(self)
        self._vip_loop_running = True

        while True:
            try:
                frames = self.zmq_router.socket.recv_multipart(copy=False)
                frames = deserialize_frames(frames)
                sender, recipient, proto, auth_token, msg_id, subsystem = frames[:6]
                sender = sender
                recipient = recipient
                subsystem = subsystem

                if subsystem == "hello":
                    self.vip.peerlist.add_peer(sender, "zmq")
                    self._zmq_peers.add(sender)
                elif subsystem == "agentstop":
                    self.vip.peerlist.drop_peer(sender, "zmq")
                    self._zmq_peers.remove(sender)
                if not recipient or recipient in self._zmq_peers:
                    # Handle router specific messages or route to ZMQ peer
                    self.zmq_router.route(frames)
                else:
                    # Route to RabbitMQ agent
                    self._route_to_agent(frames)
            except ZMQError as exc:
                # _log.error("Error while receiving message: {}".format(exc))
                if exc.errno == ENOTSOCK:
                    break
        self._vip_loop_running = False

    def _route_to_agent(self, frames):
        """
        Send the message to local agent using internal RabbitMQ message bus
        :param frames: ZMQ message frames
        :return:
        """
        sender, recipient, proto, auth_token, msg_id, subsystem = frames[:6]
        args = [arg for arg in frames[6:]]
        # for f in frames:
        #     _log.debug("Frames:; {}".format(f))
        connection = self.core.connection

        app_id = "{instance}.{identity}".format(instance=self.core.instance_name, identity=sender)
        # Change queue binding for the Response message
        # After sending the message (request) on behalf of ZMQ client, the response has to
        # routed back to the caller. Queue binding is modified for that purpose.
        # outbound_response_handler() gets called (based on the binding) to reformat response
        # message and send over zmq bus
        connection.channel.queue_bind(
            exchange=connection.exchange,
            queue=self._outbound_response_queue,
            routing_key=app_id,
            callback=None,
        )

        # Set the destination routing key to destination agent
        destination_routing_key = "{0}.{1}".format(self.core.instance_name, recipient)

        # Fit VIP frames into the PIKA properties dictionary
        # VIP format - [SENDER, RECIPIENT, PROTO, USER_ID, MSG_ID, SUBSYS, ARGS...]
        dct = {
            "user_id":
            self.core.instance_name + "." + self.core.identity,
            "app_id":
            app_id,    # Routing key of SOURCE AGENT
            "headers":
            dict(
                sender=sender,    # SENDER
                recipient=destination_routing_key,    # RECEIVER
                proto="VIP",    # PROTO
                user=auth_token,    # USER_ID
            ),
            "message_id":
            msg_id,    # MSG_ID
            "type":
            subsystem,    # SUBSYS
            "content_type":
            "application/json",
        }
        properties = pika.BasicProperties(**dct)
        # _log.debug("PROXY PUBLISHING TO CHANNEL {0}, {1}, {2}".format(destination_routing_key, app_id, properties))
        connection.channel.basic_publish(
            connection.exchange,
            destination_routing_key,
            jsonapi.dumps(args, ensure_ascii=False),
            properties,
        )



# # -*- coding: utf-8 -*- {{{
# # vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
# #
# # Copyright 2020, Battelle Memorial Institute.
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# # http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
# #
# # This material was prepared as an account of work sponsored by an agency of
# # the United States Government. Neither the United States Government nor the
# # United States Department of Energy, nor Battelle, nor any of their
# # employees, nor any jurisdiction or organization that has cooperated in the
# # development of these materials, makes any warranty, express or
# # implied, or assumes any legal liability or responsibility for the accuracy,
# # completeness, or usefulness or any information, apparatus, product,
# # software, or process disclosed, or represents that its use would not infringe
# # privately owned rights. Reference herein to any specific commercial product,
# # process, or service by trade name, trademark, manufacturer, or otherwise
# # does not necessarily constitute or imply its endorsement, recommendation, or
# # favoring by the United States Government or any agency thereof, or
# # Battelle Memorial Institute. The views and opinions of authors expressed
# # herein do not necessarily state or reflect those of the
# # United States Government or any agency thereof.
# #
# # PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# # BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# # under Contract DE-AC05-76RL01830
# # }}}
#
#

# TODO Add this module back in with plugin architecture.
# from __future__ import absolute_import
#
# import inspect
# import logging
# import uuid
# import weakref
#
# import errno
# from .base import SubsystemBase
#
# from collections import defaultdict
#
# import requests
# from requests.packages.urllib3.connection import ConnectionError, NewConnectionError
#
# from volttron.utils import ClientContext as cc
#
# from volttron.utils import jsonapi
# from ..decorators import annotate, annotations, dualmethod, spawn
# from ..errors import Unreachable
# from ..results import ResultsDictionary
#
# if cc.is_rabbitmq_available():
#     import pika
#
#
# __all__ = ["RMQPubSub"]
# min_compatible_version = "5.0"
# max_compatible_version = ""
#
#
# class RMQPubSub(SubsystemBase):
#     """
#     Pubsub subsystem concrete class implementation for RabbitMQ message bus.
#     """
#
#     def __init__(self, core, rpc_subsys, peerlist_subsys, owner):
#         self.core = weakref.ref(core)
#         self.rpc = weakref.ref(rpc_subsys)
#         self.peerlist = weakref.ref(peerlist_subsys)
#         self._owner = owner
#         self._logger = logging.getLogger(__name__)
#         self._results = ResultsDictionary()
#         self._message_number = 0
#         self._pubcount = dict()
#         self._isconnected = False
#
#         def subscriptions():
#             return defaultdict(set)
#
#         self._my_subscriptions = defaultdict(subscriptions)
#
#         def setup(sender, **kwargs):
#             # pylint: disable=unused-argument
#             core.onconnected.connect(self._connected)
#
#             def subscribe(member):  # pylint: disable=redefined-outer-name
#                 for peer, bus, prefix, all_platforms, queue_name in annotations(
#                     member, set, "pubsub.subscriptions"
#                 ):
#                     self._logger.debug("peer: {0}, prefix:{1}".format(peer, prefix))
#                     routing_key = self._form_routing_key(
#                         prefix, all_platforms=all_platforms
#                     )
#                     # If named queue, add "persistent" in the queue name
#                     if queue_name:
#                         queue_name = "{user}.pubsub.persistent.{queue_name}".format(
#                             user=self.core().rmq_user, queue_name=queue_name
#                         )
#                     else:
#                         queue_name = "{user}.pubsub.{uid}".format(
#                             user=self.core().rmq_user, uid=uuid.uuid4()
#                         )
#
#                     self._add_subscription(routing_key, member, queue_name)
#                     # self._logger.debug("SYNC RMQ: all_platforms {}")
#
#             inspect.getmembers(owner, subscribe)
#
#         core.onsetup.connect(setup, self)
#
#     def _connected(self, sender, **kwargs):
#         """
#         After connection to RMQ broker is established, synchronize local subscriptions with RMQ broker.
#         param sender: identity of sender
#         type sender: str
#         param kwargs: optional arguments
#         type kwargs: pointer to arguments
#         """
#         # self.core().connection.channel.confirm_delivery(self.on_delivery_confirmation,nowait=True)
#         self._isconnected = True
#         self.synchronize()
#
#     def synchronize(self):
#         """
#         Synchronize local subscriptions with RMQ broker.
#         :return:
#         """
#         connection = self.core().connection
#         # self._logger.debug("Synchronize {}".format(self._my_subscriptions))
#         for prefix, subscriptions in self._my_subscriptions.items():
#             for queue_name, callback in subscriptions.items():
#                 durable = False
#                 auto_delete = True
#                 # Check if queue needs to be persistent
#                 if "persistent" in queue_name:
#                     durable = True
#                     auto_delete = False
#                 connection.channel.queue_declare(
#                     queue=queue_name,
#                     durable=durable,
#                     exclusive=True,
#                     auto_delete=auto_delete,
#                     callback=None,
#                 )
#                 connection.channel.queue_bind(
#                     exchange=connection.exchange,
#                     queue=queue_name,
#                     routing_key=prefix,
#                     callback=None,
#                 )
#
#                 if prefix.startswith("__pubsub__.*."):
#                     original_prefix = self._get_original_topic(prefix)
#                     self._send_proxy(original_prefix)
#                 for cb in callback:
#                     self._add_callback(connection, queue_name, cb)
#         return True
#
#     def _add_subscription(self, prefix, callback, queue_name):
#         """
#         Store subscriptions so that it can used later
#         :param prefix: subscription prefix
#         :param callback: callback method
#         :param queue: queue name
#         :return:
#         """
#         if not callable(callback):
#             raise ValueError("callback %r is not callable" % (callback,))
#         try:
#             self._my_subscriptions[prefix][queue_name].add(callback)
#             # _log.debug("SYNC: add subscriptions: {}".format(self._my_subscriptions['internal'][bus][prefix]))
#         except KeyError:
#             self._logger.error("PUBSUB something went wrong when adding subscriptions")
#
#     @dualmethod
#     @spawn
#     def subscribe(
#         self, peer, prefix, callback, bus="", all_platforms=False, persistent_queue=None
#     ):
#         """Subscribe to a prefix and register callback. If 'all_platforms' flag is set to True, then
#         agent subscribes to receive topic from all platforms. A named queue will set persistent
#         behavior to the topic subscriptions. That means even if the agent shutdowns and restarts, it
#         will receive all the messages during the shutdown/turn off period.
#
#         :param peer "pubsub" string
#         :type peer str
#         :param prefix prefix of the topic
#         :type prefix str
#         :param callback callback method
#         :type callback method
#         :param bus bus
#         :type bus str
#         :param all_platforms Flag indicating if type is 'local' or 'all'
#         :type all_platforms boolean
#         :param persistent_queue Name of the queue for persistent behavior
#         :type persistent_queue str
#         :returns: Subscribe is successful or not
#         :rtype: boolean
#
#         :Return Values:
#         Success or Failure
#         """
#         result = None
#         connection = self.core().connection  # bytes(uuid.uuid4())
#         routing_key = self._form_routing_key(prefix, all_platforms=all_platforms)
#         if all_platforms:
#             # Send message to proxy agent in order to subscribe with zmq message bus
#             self._send_proxy(prefix)
#
#         queue_name = ""
#         durable = False
#         auto_delete = True
#
#         if persistent_queue:
#             durable = True
#             auto_delete = False
#             queue_name = "{user}.pubsub.persistent.{queue_name}".format(
#                 user=self.core().rmq_user, queue_name=persistent_queue
#             )
#         else:
#             queue_name = "{user}.pubsub.{uid}".format(
#                 user=self.core().rmq_user, uid=str(uuid.uuid4())
#             )
#         # Store subscriptions for later use
#         self._add_subscription(routing_key, callback, queue_name)
#
#         self._logger.debug("RMQ PUBSUB subscribing to {}".format(routing_key))
#
#         try:
#             connection.channel.queue_declare(
#                 callback=None,
#                 queue=queue_name,
#                 durable=durable,
#                 exclusive=False,
#                 auto_delete=auto_delete,
#             )
#             connection.channel.queue_bind(
#                 callback=None,
#                 exchange=connection.exchange,
#                 queue=queue_name,
#                 routing_key=routing_key,
#             )
#             self._add_callback(connection, queue_name, callback)
#         except AttributeError as ex:
#             self._logger.error(
#                 "Subscription will be added when agent gets connected to messagebus.".format(
#                     self.core().identity
#                 )
#             )
#         return result
#
#     def _send_proxy(self, prefix, bus=""):
#         """
#         Send the message to proxy router
#         :param prefix:
#         :param bus:
#         :return:
#         """
#         connection = self.core().connection
#         rkey = self.core().instance_name + ".proxy.router.pubsub"
#         sub_msg = dict(prefix=prefix, bus=bus, all_platforms=True)
#         # VIP format - [SENDER, RECIPIENT, PROTO, USER_ID, MSG_ID, SUBSYS, ARGS...]
#         frames = [
#             self.core().identity,
#             "",
#             "VIP1",
#             "",
#             "",
#             "pubsub",
#             "subscribe",
#             sub_msg,
#         ]
#         connection.channel.basic_publish(
#             exchange=connection.exchange,
#             routing_key=rkey,
#             body=jsonapi.dumps(frames, ensure_ascii=False),
#         )
#
#     def _add_callback(self, connection, queue, callback):
#         """
#         Register agent's callback method with RabbitMQ broker
#         :param connection: RabbitMQ connection object
#         :param queue: queue name
#         :param callback: callback method
#         :return:
#         """
#
#         def rmq_callback(ch, method, properties, body):
#             # Strip prefix from routing key
#             topic = self._get_original_topic(str(method.routing_key))
#             try:
#                 msg = jsonapi.loads(body)
#                 headers = msg["headers"]
#                 message = msg["message"]
#                 bus = msg["bus"]
#                 sender = msg["sender"]
#                 self.core().spawn(
#                     callback, "pubsub", sender, bus, topic, headers, message
#                 )
#             except KeyError as esc:
#                 self._logger.error("Missing keys in pubsub message {}".format(esc))
#
#         connection.channel.basic_consume(rmq_callback, queue=queue, no_ack=True)
#
#     @subscribe.classmethod
#     def subscribe(
#         cls, peer, prefix, bus="", all_platforms=False, persistent_queue=None
#     ):
#         """
#         Class method for subscribe
#         :param peer: "pubsub" string
#         :param prefix: prefix of the topic
#         :param bus: bus
#         :param all_platforms: Flag indicating if type is 'local' or 'all'
#         :param persistent_queue: Name of the queue for persistent behavior
#         :return:
#         """
#
#         def decorate(method):
#             annotate(
#                 method,
#                 set,
#                 "pubsub.subscriptions",
#                 (peer, bus, prefix, all_platforms, persistent_queue),
#             )
#             return method
#
#         return decorate
#
#     def list(
#         self,
#         peer,
#         prefix="",
#         bus="",
#         subscribed=True,
#         reverse=False,
#         all_platforms=False,
#     ):
#         """Gets list of subscriptions matching the prefix
#         param peer: peer
#         type peer: str
#         param prefix: prefix of a topic
#         type prefix: str
#         param bus: bus
#         type bus: bus
#         param subscribed: subscribed or not
#         type subscribed: boolean
#         param reverse: reverse
#         type reverse:
#         :returns: List of subscriptions, i.e, list of tuples of bus, topic and flag to indicate if peer is a
#         subscriber or not
#         :rtype: list of tuples
#
#         :Return Values:
#         List of tuples [(bus, topic, flag to indicate if peer is a subscriber or not)]
#         """
#         async_result = next(self._results)
#         results = []
#         if reverse:
#             test = prefix.startswith
#         else:
#             test = lambda t: t.startswith(prefix)
#
#         try:
#             bindings = self.core().rmq_mgmt.get_bindings("volttron")
#         except (
#             requests.exceptions.HTTPError,
#             ConnectionError,
#             NewConnectionError,
#         ) as e:
#             self._logger.error(
#                 "Error making request to RabbitMQ Management interface.\n"
#                 "Check Connection Parameters: {} \n".format(e)
#             )
#         else:
#             try:
#                 items = [
#                     (b["destination"], self._get_original_topic(b["routing_key"]))
#                     for b in bindings
#                     if b["routing_key"].startswith("__pubsub__")
#                 ]
#             except KeyError as e:
#                 return async_result
#
#             for item in items:
#                 peer = item[0]
#                 topic = item[1]
#                 if test(topic):
#                     member = self.core().identity in peer
#                     if not subscribed or member:
#                         results.append(("", topic, member))
#         self.core().spawn_later(0.01, self.set_result, async_result.ident, results)
#         return async_result
#
#     def publish(self, peer, topic, headers=None, message=None, bus=""):
#         """Publish a message to a given topic via a peer.
#
#         Publish headers and message to all subscribers of topic on bus.
#         If peer is None, use self. Adds volttron platform version
#         compatibility information to header as variables
#         min_compatible_version and max_compatible version
#         param peer: peer
#         type peer: str
#         param topic: topic for the publish message
#         type topic: str
#         param headers: header info for the message
#         type headers: None or dict
#         param message: actual message
#         type message: None or any
#         param bus: bus
#         type bus: str
#         return: Number of subscribers the message was sent to.
#         :rtype: int
#
#         :Return Values:
#         Number of subscribers
#         """
#         result = next(self._results)
#         self._pubcount[self._message_number] = result.ident
#         self._message_number += 1
#         routing_key = self._form_routing_key(topic)
#         connection = self.core().connection
#         self.core().spawn_later(0.01, self.set_result, result.ident, 1)
#         if headers is None:
#             headers = {}
#
#         headers["min_compatible_version"] = min_compatible_version
#         headers["max_compatible_version"] = max_compatible_version
#         # self._logger.debug("RMQ PUBSUB publish message To. {0}, {1}, {2}, {3} ".format(routing_key,
#         #                                                                            self.core().identity,
#         #                                                                            message,
#         #                                                                            topic))
#
#         # VIP format - [SENDER, RECIPIENT, PROTO, USER_ID, MSG_ID, SUBSYS, ARGS...]
#         dct = {
#             # 'user_id': self.core().identity,
#             "app_id": connection.routing_key,  # SENDER
#             "headers": dict(
#                 recipient="",  # RECEIVER
#                 proto="VIP",  # PROTO
#                 user=self.core().identity,  # USER_ID
#             ),
#             "message_id": result.ident,  # MSG_ID
#             "type": "pubsub",  # SUBSYS
#             "content_type": "application/json",
#         }
#         properties = pika.BasicProperties(**dct)
#         json_msg = dict(
#             sender=self.core().identity, bus=bus, headers=headers, message=message
#         )
#         try:
#             connection.channel.basic_publish(
#                 exchange=connection.exchange,
#                 routing_key=routing_key,
#                 properties=properties,
#                 body=jsonapi.dumps(json_msg, ensure_ascii=False),
#             )
#         except (
#             pika.exceptions.AMQPConnectionError,
#             pika.exceptions.AMQPChannelError,
#         ) as exc:
#             self._isconnected = False
#             raise Unreachable(
#                 errno.EHOSTUNREACH,
#                 "Connection to RabbitMQ is lost",
#                 "rabbitmq broker",
#                 "pubsub",
#             )
#         return result
#
#     def set_result(self, ident, value=None):
#         try:
#             result = self._results.pop(ident)
#             if result:
#                 result.set(value)
#         except KeyError:
#             pass
#
#     def on_delivery_confirmation(self, method_frame):
#         """Invoked by pika when RabbitMQ responds to a Basic.Publish RPC
#         command, passing in either a Basic.Ack or Basic.Nack frame with
#         the delivery tag of the message that was published. The delivery tag
#         is an integer counter indicating the message number that was sent
#         on the channel via Basic.Publish. Here we're just doing house keeping
#         to keep track of stats and remove message numbers that we expect
#         a delivery confirmation of from the list used to keep track of messages
#         that are pending confirmation.
#
#         :param pika.frame.Method method_frame: Basic.Ack or Basic.Nack frame
#
#         """
#         try:
#             delivery_number = method_frame.method.delivery_tag
#             self._logger.info(
#                 "PUBSUB Delivery confirmation {0}, pending {1}, ".format(
#                     method_frame.method.delivery_tag, len(self._pubcount)
#                 )
#             )
#             ident = self._pubcount.pop(delivery_number, None)
#             if ident:
#                 result = None
#                 try:
#                     result = self._results.pop(ident)
#                     if result:
#                         result.set(delivery_number)
#                 except KeyError:
#                     pass
#         except TypeError:
#             pass
#
#     def unsubscribe(self, peer, prefix, callback, bus="", all_platforms=False):
#         """Unsubscribe and remove callback(s).
#
#         Remove all handlers matching the given info - peer, callback and bus,
#         which was used earlier to subscribe as well. If all handlers for a
#         topic prefix are removed, the topic is also unsubscribed.
#         param peer: peer
#         type peer: str
#         param prefix: prefix that needs to be unsubscribed
#         type prefix: str
#         param callback: callback method
#         type callback: method
#         param bus: bus
#         type bus: bus
#         return: success or not
#         :rtype: boolean
#
#         :Return Values:
#         success or not
#         """
#         routing_key = None
#
#         result = next(self._results)
#         if prefix is not None:
#             routing_key = self._form_routing_key(prefix, all_platforms=all_platforms)
#         topics = self._drop_subscription(routing_key, callback)
#         self.core().spawn_later(0.01, self.set_result, result.ident, topics)
#         # Send the message to proxy router to send it to external 'zmq' platforms
#         if all_platforms:
#             subscriptions = dict()
#             subscriptions["all"] = dict(prefix=topics, bus=bus)
#             rkey = self.core().instance_name + ".proxy.router.pubsub"
#             frames = [
#                 self.core().identity,
#                 "",
#                 "VIP1",
#                 "",
#                 "",
#                 "pubsub",
#                 "unsubscribe",
#                 jsonapi.dumps(subscriptions),
#             ]
#             self.core().connection.channel.basic_publish(
#                 exchange=self.core().connection.exchange, routing_key=rkey, body=frames
#             )
#         return result
#
#     def _drop_subscription(self, routing_key, callback):
#         """
#         Utility method to remove subscription
#         :param routing_key: routing key
#         :param callback: callback method
#         :return:
#         """
#         self._logger.debug("DROP subscriptions: {}".format(routing_key))
#         topics = []
#         remove = []
#         remove_topics = []
#         if routing_key is None:
#             if callback is None:
#                 for prefix in self._my_subscriptions:
#                     subscriptions = self._my_subscriptions[prefix]
#                     for queue_name in list(subscriptions):
#                         self.core().connection.channel.queue_delete(
#                             callback=None, queue=queue_name
#                         )
#                         subscriptions.pop(queue_name)
#                     topics.append(prefix)
#             else:
#                 # Traverse through all subscriptions to find the callback
#                 for prefix in self._my_subscriptions:
#                     subscriptions = self._my_subscriptions[prefix]
#                     self._logger.debug("prefix: {0}, {1}".format(prefix, subscriptions))
#                     for queue_name, callbacks in subscriptions.items():
#                         try:
#                             callbacks.remove(callback)
#                         except KeyError:
#                             pass
#                         else:
#                             topics.append(prefix)
#                         if not callbacks:
#                             # Delete queue
#                             self.core().connection.channel.queue_delete(
#                                 callback=None, queue=queue_name
#                             )
#                             remove.append(queue_name)
#                     for que in remove:
#                         del subscriptions[que]
#                     del remove[:]
#                     if not subscriptions:
#                         remove_topics.append(prefix)
#                 for prefix in remove_topics:
#                     del self._my_subscriptions[prefix]
#                 if not topics:
#                     raise KeyError("no such subscription")
#                 self._logger.debug(
#                     "my subscriptions: {0}".format(self._my_subscriptions)
#                 )
#         else:
#             # Search based on routing key
#             if routing_key in self._my_subscriptions:
#                 self._logger.debug(
#                     "RMQ subscriptions {}".format(self._my_subscriptions)
#                 )
#                 topics.append(routing_key)
#                 subscriptions = self._my_subscriptions[routing_key]
#                 if callback is None:
#                     for queue_name, callbacks in subscriptions.items():
#                         self._logger.debug("RMQ queues {}".format(queue_name))
#                         self.core().connection.channel.queue_delete(
#                             callback=None, queue=queue_name
#                         )
#                     del self._my_subscriptions[routing_key]
#                 else:
#                     self._logger.debug("topics: {0}".format(topics))
#                     for queue_name, callbacks in subscriptions.items():
#                         try:
#                             callbacks.remove(callback)
#                         except KeyError:
#                             pass
#                         if not callbacks:
#                             # Delete queue
#                             self.core().connection.channel.queue_delete(
#                                 callback=None, queue=queue_name
#                             )
#                             remove.append(queue_name)
#                     for que in remove:
#                         del subscriptions[que]
#                     if not subscriptions:
#                         del self._my_subscriptions[routing_key]
#             self._logger.debug("my subscriptions: {0}".format(self._my_subscriptions))
#         orig_topics = []
#         # Strip '__pubsub__.<instance_name>' from the topic string
#         for topic in topics:
#             orig_topics.append(self._get_original_topic(topic))
#         # self._logger.debug("AFTER DROP topics: {}".format(orig_topics))
#         return orig_topics
#
#     def _get_original_topic(self, routing_key):
#         """
#         Replace '.' delimiter with '/'
#         :param routing_key: routing_key string
#         :return: return original topic string
#         """
#         try:
#             original_topic = routing_key.split(".")[2:]
#             original_topic = original_topic[:-1]
#             original_topic = "/".join(original_topic)
#             return original_topic
#         except IndexError as exc:
#             return routing_key
#
#     def _form_routing_key(self, topic, all_platforms=False):
#         """
#         Form routing key from the original topic
#         :param topic: Original topic
#         :param all_platforms: Flag indicating if it is intended for all platforms
#         :return: Routing key string
#         """
#         routing_key = ""
#         topic = "#" if topic == "" else topic + ".#"
#
#         if all_platforms:
#             # Format is '__pubsub__.*.<prefix>.#'
#             routing_key = "__pubsub__.*.{}".format(topic.replace("/", "."))
#         else:
#             routing_key = "__pubsub__.{0}.{1}".format(
#                 self.core().instance_name, topic.replace("/", ".")
#             )
#         return routing_key



import threading


class Monitor(threading.Thread):
    """Monitor thread to log connections."""

    def __init__(self, sock):
        super(Monitor, self).__init__()
        self.daemon = True
        self.sock = sock

    def run(self):
        events = {
            value: name[6:]
            for name, value in vars(zmq).items()
            if name.startswith("EVENT_") and name != "EVENT_ALL"
        }
        log = logging.getLogger("vip.monitor")
        if log.level == logging.NOTSET:
            log.setLevel(logging.INFO)
        sock = self.sock
        while True:
            event, endpoint = sock.recv_multipart()
            event_id, event_value = struct.unpack("=HI", event)
            event_name = events[event_id]
            log.info("%s %s %s", event_name, event_value, endpoint)
